""" @module
This module contains the admitted diagnostics.
Diagnostics are stored in a dictionary. Keys are literal descriptions of the diagnostic 
(e.g. '[NI] 5198/5200') and items are tuples including a label ('N1'), an expression for the line ratio ('L(5198)/L(5200)',
and an expression for the uncertainty on the line ratio as a function of the uncertainty on the individual lines ('RMS([E(5200), E(5198)])')

"""
import numpy as np
import pyneb as pn
from pyneb.utils.misc import int_to_roman, parseAtom, parseAtom2
from pyneb.utils.init import BLEND_LIST

diags_dict = {}

diags_dict['[CIII] 1909/1907'] = ('C3', 'L(1909)/L(1907)', 'RMS([E(1909), E(1907)])')
diags_dict['[NI] 5198/5200'] = ('N1', 'I(3, 1)/I(2, 1)', 'RMS([E(5200), E(5198)])')
diags_dict['[NII] 5755/6548'] = ('N2', 'L(5755)/L(6548)', 'RMS([E(6548), E(5755)])')
diags_dict['[NII] 5755/6584'] = ('N2', 'L(5755)/L(6584)', 'RMS([E(6584), E(5755)])')
diags_dict['[NII] 5755/6584+'] = ('N2', 'L(5755)/(L(6548)+L(6584))', 'RMS([E(6548)*L(6548)/(L(6548)+L(6584)), E(6584)*L(6584)/(L(6584)+L(6548)), E(5755)])')
diags_dict['[NII] 121m/20.5m'] = ('N2', 'L(1214747)/L(2054427)', 'RMS([E(2054427)/E(1214747)])')
diags_dict['[NIII] 1751+/57.4m'] = ('N3', 'B("1751A+")/L(574000)', 'RMS([E(574000), BE("1751A+")])')
#diags_dict[''] = ('N3','L(1749)/L(1752)','RMS([E(1749), E(1752)])')
diags_dict['[OI] 63m/147m'] = ('O1', 'L(632000)/L(1455000)', 'RMS([E(632000), E(1455000)])')
diags_dict['[OI] 5577/6302'] = ('O1', 'L(5577)/L(6300)', 'RMS([E(6300), E(5577)])')
diags_dict['[OI] 5577/6300'] = ('O1', 'L(5577)/L(6300)', 'RMS([E(6300), E(5577)])')
diags_dict['[OI] 5577/6300+'] = ('O1', 'L(5577)/(L(6300)+L(6364))', 'RMS([E(6300)*L(6300)/(L(6300)+L(6364)), E(6364)*L(6364)/(L(6300)+L(6364)), E(5577)])')
diags_dict['[OII] 3726/3729'] = ('O2', 'L(3726)/L(3729)', 'RMS([E(3729), E(3726)])')
diags_dict['[OII] 3727+/7325+c'] =  ('O2', '(B("3727A+"))/(B("7319A+")+B("7330A+"))',
            'RMS([BE("7319A+")*B("7319A+")/(B("7319A+")+B("7330A+")), BE("7330A+")*B("7330A+")/(B("7319A+")+B("7330A+")), BE("3727A+")])')
diags_dict['[OII] 3727+/7325+'] = ('O2', '(L(3726)+L(3729))/(B("7319A+")+B("7330A+"))',
              'RMS([E(3726)*L(3726)/(L(3726)+L(3729)), E(3729)*L(3729)/(L(3726)+L(3729)),BE("7319A+")*B("7319A+")/(B("7319A+")+B("7330A+")),BE("7330A+")*B("7330A+")/(B("7319A+")+B("7330A+"))])')
#diags_dict['[OII] 3727+/7325+'] = ('O2', '(L(3726)+L(3729))/(B("7325A+"))', 'RMS([E(3726)*L(3726)/(L(3726)+L(3729)), E(3729)*L(3729)/(L(3726)+L(3729)),BE("7325A+")])')
diags_dict['[OII] 3727+/7325+b'] = ('O2', '(L(3726)+L(3729))/(I(4,2)+I(4,3)+I(5,2)+I(5,3))',
              'RMS([E(3726)*L(3726)/(L(3726)+L(3729)), E(3729)*L(3729)/(L(3726)+L(3729)),BE("7319A+")*B("7319A+")/(B("7319A+")+B("7330A+")),BE("7330A+")*B("7330A+")/(B("7319A+")+B("7330A+"))])')
diags_dict['OII 4649.13/4089.29'] = ('O2r', "S('4649.13')/S('4089.29')", "RMS([SE('4649.13'), SE('4089.29')])")
diags_dict['OII 4649.13/4661.63'] = ('O2r', "S('4649.13')/S('4661.63')", "RMS([SE('4649.13'), SE('4661.63')])")
diags_dict['[OIII] 4363/5007'] = ('O3', 'L(4363)/L(5007)', 'RMS([E(5007), E(4363)])')
diags_dict['[OIII] 4363/5007+'] = ('O3', 'L(4363)/(L(5007)+L(4959))', 'RMS([E(5007)*L(5007)/(L(5007)+L(4959)), E(4959)*L(4959)/(L(5007)+L(4959)), E(4363)])')
diags_dict['[OIII] 5007/88m'] = ('O3', 'L(5007)/L(883000)', 'RMS([E(883000), E(5007)])')
diags_dict['[OIII] 51m/88m'] = ('O3', 'L(518000)/L(883000)', 'RMS([E(883000), E(518000)])')
diags_dict['[OIII] 1666/5007+'] = ('O3', 'L(1666)/(L(5007)+L(4959))', 'RMS([E(5007)*L(5007)/(L(5007)+L(4959)), E(4959)*L(4959)/(L(5007)+L(4959)), E(1666)])')
diags_dict['[OIII] 1666/5007'] = ('O3', 'L(1666)/L(5007)', 'RMS([E(5007), E(1666)])')
diags_dict['[OIII] 1664+/5007'] = ('O3', '(B("1664A+"))/L(5007)', 'RMS([BE("1664A+"), E(5007)])')
diags_dict['[OIII] 1666/4363'] = ('O3', 'L(1666)/L(4363)', 'RMS([E(4363), E(1666)])')
diags_dict['[OIV] 1401/1405'] = ('O4', 'L(1401)/L(1405)', 'RMS([E(1401), E(1405)])')
diags_dict['[OIV] 1400+/25.9m'] = ('O4', 'B("1400A+")/L(259000)', 'RMS([BE("1400A+"), E(259000)])')
diags_dict['[SII] 6731/6716'] = ('S2', 'L(6731)/L(6716)', 'RMS([E(6716), E(6731)])')
diags_dict['[SII] 4069/4076'] = ('S2', 'L(4069)/L(4076)', 'RMS([E(4069), E(4076)])')
diags_dict['[SII] 4072+/6720+'] = ('S2', '(L(4069)+L(4076))/(L(6716)+L(6731))',
              'RMS([E(6716)*L(6716)/(L(6716)+L(6731)), E(6731)*L(6731)/(L(6716)+L(6731)), E(4069)*L(4069)/(L(4069)+L(4076)), E(4076)*L(4076)/(L(4069)+L(4076))])')
diags_dict['[SIII] 18.7m/33.5m'] = ('S3', 'L(187000)/L(335000)', 'RMS([E(335000), E(187000)])')
diags_dict['[SIII] 6312/18.7m'] = ('S3', 'L(6312)/L(187000)', 'RMS([E(187000), E(6312)])')
diags_dict['[SIII] 9069/18.7m'] = ('S3', 'L(9069)/L(187000)', 'RMS([E(187000), E(9069)])')
diags_dict['[SIII] 6312/9200+'] = ('S3', 'L(6312)/(L(9069)+L(9531))', 'RMS([E(9069)*L(9069)/(L(9069)+L(9531)), E(9531)*L(9531)/(L(9069)+L(9531)), E(6312)])')
diags_dict['[SIII] 6312/9069'] = ('S3', 'L(6312)/L(9069)', 'RMS([E(9069), E(6312)])')
diags_dict['[NeIII] 15.6m/36.0m'] = ('Ne3', 'L(156000)/L(360000)', 'RMS([E(156000), E(360000)])')
diags_dict['[NeIII] 3869/15.6m'] = ('Ne3', 'L(3869)/L(156000)', 'RMS([E(156000), E(3869)])')
diags_dict['[NeIII] 3930+/15.6m'] = ('Ne3', '(L(3869)+L(3968))/L(156000)', 'RMS([E(156000), E(3869)*L(3869)/(L(3869)+L(3968)), E(3968)*L(3968)/(L(3869)+L(3968))])')
diags_dict['[NeIII] 3344/3930+'] = ('Ne3', 'L(3343)/(L(3869)+L(3968))', 'RMS([E(3869)*L(3869)/(L(3869)+L(3968)), E(3968)*L(3968)/(L(3869)+L(3968)), E(3343)])')
diags_dict['[NeIII] 3343/3930+'] = ('Ne3', 'L(3343)/(L(3869)+L(3968))', 'RMS([E(3869)*L(3869)/(L(3869)+L(3968)), E(3968)*L(3968)/(L(3869)+L(3968)), E(3343)])')
diags_dict['[NeV] 2973/3370+'] = ('Ne5', 'L(2973)/(L(3426)+L(3346))', 'RMS([E(3426)*L(3426)/(L(3426)+L(3346)), E(3346)*L(3346)/(L(3426)+L(3346)), E(2973)])')
diags_dict['[NeV] 1575/3426'] = ('Ne5', 'L(1575)/L(3426)', 'RMS([E(1575), E(3426)])')
diags_dict['[NeV] 14.3m/24.2m'] = ('Ne5', 'L(143000)/L(242000)', 'RMS([E(143000), E(242000)])')
diags_dict['[NeV] 3426/24.2m'] = ('Ne5', 'L(3426)/L(242000)', 'RMS([E(3426), E(242000)])')
diags_dict['[ClIII] 5538/5518'] = ('Cl3', 'L(5538)/L(5518)', 'RMS([E(5518), E(5538)])')
diags_dict['[ClIV] 5323/7531'] = ('Cl4', 'L(5323)/L(7531)', 'RMS([E(7531), E(5323)])')
diags_dict['[ClIV] 5323/7700+'] = ('Cl4', 'L(5323)/(L(7531)+L(8046))', 'RMS([E(7531)*L(7531)/(L(7531)+L(8046)), E(8046)*L(8046)/(L(7531)+L(8046)), E(5323)])')
diags_dict['[ArIII] (7751+7136)/9m'] = ('Ar3', '(L(7751)+L(7136))/L(90000)', 'RMS([E(90000), E(7751)*L(7751)/(L(7751)+L(7136)), E(7136)*L(7136)/(L(7751)+L(7136))])')
diags_dict['[ArIII] 7136/9m'] = ('Ar3', 'L(7136)/L(90000)', 'RMS([E(90000), E(7136)])')
diags_dict['[ArIII] 5192/7300+'] = ('Ar3', 'L(5192)/(L(7751)+L(7136))', 'RMS([E(7751)*L(7751)/(L(7751)+L(7136)), E(7136)*L(7136)/(L(7751)+L(7136)), E(5192)])')
diags_dict['[ArIII] 5192/7136'] = ('Ar3', 'L(5192)/L(7136)', 'RMS([E(7136), E(5192)])')
diags_dict['[ArIII] 9.0m/21.8m'] = ('Ar3', 'L(89897)/L(218000)', 'RMS([E(89897), E(218000)])')
diags_dict['[ArIV] 4740/4711'] = ('Ar4', 'L(4740)/L(4711)', 'RMS([E(4711), E(4740)])')
diags_dict['[ArIV] 2860+/4720+'] = ('Ar4', '(L(2854)+L(2868))/(L(4711)+L(4740))',
              'RMS([E(4711)*L(4711)/(L(4711)+L(4740)), E(4740)*L(4740)/(L(4711)+L(4740)), E(2854)*L(2854)/(L(2854)+L(2868)), E(2868)*L(2854)/(L(2854)+L(2868))])')
diags_dict['[ArIV] 7230+/4720+'] = ('Ar4', '(L(7170)+L(7263))/(L(4711)+L(4740))',
              'RMS([E(4711)*L(4711)/(L(4711)+L(4740)), E(4740)*L(4740)/(L(4711)+L(4740)), E(7170)*L(7170)/(L(7170)+L(7263)), E(7263)*L(7263)/(L(7170)+L(7263))])')
diags_dict['[ArV] 4626/6600+'] = ('Ar5', 'L(4626)/(L(6435)+L(7005))', 'RMS([E(6435)*L(6435)/(L(6435)+L(7005)), E(7005)*L(7005)/(L(6435)+L(7005)), E(4626)])')

diags_dict['[FeIII] 5270/4987'] = ('Fe3', 'L(5270)/L(4987)', 'RMS([E(5270), E(4987)])')
diags_dict['[FeIII] 5270/4925'] = ('Fe3', 'L(5270)/L(4925)', 'RMS([E(5270), E(4925)])')
diags_dict['[FeIII] 5270/4881'] = ('Fe3', 'L(5270)/L(4881)', 'RMS([E(5270), E(4881)])')
diags_dict['[FeIII] 5270/5011'] = ('Fe3', 'L(5270)/L(5011)', 'RMS([E(5270), E(5011)])')
diags_dict['[FeIII] 5270/4931'] = ('Fe3', 'L(5270)/L(4931)', 'RMS([E(5270), E(4931)])')
diags_dict['[FeIII] 5270/4659'] = ('Fe3', 'L(5270)/L(4659)', 'RMS([E(5270), E(4659)])')
diags_dict['[FeIII] 5270/4701'] = ('Fe3', 'L(5270)/L(4701)', 'RMS([E(5270), E(4701)])')
diags_dict['[FeIII] 5270/4734'] = ('Fe3', 'L(5270)/L(4734)', 'RMS([E(5270), E(4734)])')
diags_dict['[FeIII] 5270/4009'] = ('Fe3', 'L(5270)/L(4009)', 'RMS([E(5270), E(4009)])')
diags_dict['[FeIII] 4987/4925'] = ('Fe3', 'L(4987)/L(4925)', 'RMS([E(4987), E(4925)])')
diags_dict['[FeIII] 4987/4881'] = ('Fe3', 'L(4987)/L(4881)', 'RMS([E(4987), E(4881)])')
diags_dict['[FeIII] 4987/5011'] = ('Fe3', 'L(4987)/L(5011)', 'RMS([E(4987), E(5011)])')
diags_dict['[FeIII] 4987/4931'] = ('Fe3', 'L(4987)/L(4931)', 'RMS([E(4987), E(4931)])')
diags_dict['[FeIII] 4987/4659'] = ('Fe3', 'L(4987)/L(4659)', 'RMS([E(4987), E(4659)])')
diags_dict['[FeIII] 4987/4701'] = ('Fe3', 'L(4987)/L(4701)', 'RMS([E(4987), E(4701)])')
diags_dict['[FeIII] 4987/4734'] = ('Fe3', 'L(4987)/L(4734)', 'RMS([E(4987), E(4734)])')
diags_dict['[FeIII] 4987/4009'] = ('Fe3', 'L(4987)/L(4009)', 'RMS([E(4987), E(4009)])')
diags_dict['[FeIII] 4925/4881'] = ('Fe3', 'L(4925)/L(4881)', 'RMS([E(4925), E(4881)])')
diags_dict['[FeIII] 4925/5011'] = ('Fe3', 'L(4925)/L(5011)', 'RMS([E(4925), E(5011)])')
diags_dict['[FeIII] 4925/4931'] = ('Fe3', 'L(4925)/L(4931)', 'RMS([E(4925), E(4931)])')
diags_dict['[FeIII] 4925/4659'] = ('Fe3', 'L(4925)/L(4659)', 'RMS([E(4925), E(4659)])')
diags_dict['[FeIII] 4925/4701'] = ('Fe3', 'L(4925)/L(4701)', 'RMS([E(4925), E(4701)])')
diags_dict['[FeIII] 4925/4734'] = ('Fe3', 'L(4925)/L(4734)', 'RMS([E(4925), E(4734)])')
diags_dict['[FeIII] 4925/4009'] = ('Fe3', 'L(4925)/L(4009)', 'RMS([E(4925), E(4009)])')
diags_dict['[FeIII] 4881/5011'] = ('Fe3', 'L(4881)/L(5011)', 'RMS([E(4881), E(5011)])')
diags_dict['[FeIII] 4881/4931'] = ('Fe3', 'L(4881)/L(4931)', 'RMS([E(4881), E(4931)])')
diags_dict['[FeIII] 4881/4659'] = ('Fe3', 'L(4881)/L(4659)', 'RMS([E(4881), E(4659)])')
diags_dict['[FeIII] 4881/4701'] = ('Fe3', 'L(4881)/L(4701)', 'RMS([E(4881), E(4701)])')
diags_dict['[FeIII] 4881/4734'] = ('Fe3', 'L(4881)/L(4734)', 'RMS([E(4881), E(4734)])')
diags_dict['[FeIII] 4881/4009'] = ('Fe3', 'L(4881)/L(4009)', 'RMS([E(4881), E(4009)])')
diags_dict['[FeIII] 5011/4931'] = ('Fe3', 'L(5011)/L(4931)', 'RMS([E(5011), E(4931)])')
diags_dict['[FeIII] 5011/4659'] = ('Fe3', 'L(5011)/L(4659)', 'RMS([E(5011), E(4659)])')
diags_dict['[FeIII] 5011/4701'] = ('Fe3', 'L(5011)/L(4701)', 'RMS([E(5011), E(4701)])')
diags_dict['[FeIII] 5011/4734'] = ('Fe3', 'L(5011)/L(4734)', 'RMS([E(5011), E(4734)])')
diags_dict['[FeIII] 5011/4009'] = ('Fe3', 'L(5011)/L(4009)', 'RMS([E(5011), E(4009)])')
diags_dict['[FeIII] 4931/4659'] = ('Fe3', 'L(4931)/L(4659)', 'RMS([E(4931), E(4659)])')
diags_dict['[FeIII] 4931/4701'] = ('Fe3', 'L(4931)/L(4701)', 'RMS([E(4931), E(4701)])')
diags_dict['[FeIII] 4931/4734'] = ('Fe3', 'L(4931)/L(4734)', 'RMS([E(4931), E(4734)])')
diags_dict['[FeIII] 4931/4009'] = ('Fe3', 'L(4931)/L(4009)', 'RMS([E(4931), E(4009)])')
diags_dict['[FeIII] 4659/4701'] = ('Fe3', 'L(4659)/L(4701)', 'RMS([E(4659), E(4701)])')
diags_dict['[FeIII] 4659/4734'] = ('Fe3', 'L(4659)/L(4734)', 'RMS([E(4659), E(4734)])')
diags_dict['[FeIII] 4659/4009'] = ('Fe3', 'L(4659)/L(4009)', 'RMS([E(4659), E(4009)])')
diags_dict['[FeIII] 4701/4734'] = ('Fe3', 'L(4701)/L(4734)', 'RMS([E(4701), E(4734)])')
diags_dict['[FeIII] 4701/4009'] = ('Fe3', 'L(4701)/L(4009)', 'RMS([E(4701), E(4009)])')
diags_dict['[FeIII] 4734/4009'] = ('Fe3', 'L(4734)/L(4009)', 'RMS([E(4734), E(4009)])')
# To be checked:
# 4989 is to be changed into 4987, but what about 4987?
# which is the relationship between the 3 lines in the numerator?
# is the error right?
# diags_dict['[FeIII] 4659+/4987+'] = ('Fe3', '(L(4659)+L(4734)+L(4009))/(L(4987)+L(4989))', 'RMS([E(4659), E(4987)])')
diags_dict['[NiIII] 6000/7890'] = ('Ni3', 'L(6000)/L(7890)', 'RMS([E(6000), E(7890)])')
diags_dict['[NiIII] 6401/7890'] = ('Ni3', 'L(6401)/L(7890)', 'RMS([E(6401), E(7890)])')
diags_dict['[NiIII] 6534/7890'] = ('Ni3', 'L(6534)/L(7890)', 'RMS([E(6534), E(7890)])')
diags_dict['[NiIII] 6682/7890'] = ('Ni3', 'L(6682)/L(7890)', 'RMS([E(6682), E(7890)])')
diags_dict['[NiIII] 6797/7890'] = ('Ni3', 'L(6797)/L(7890)', 'RMS([E(6797), E(7890)])')
diags_dict['[NiIII] 6946/7890'] = ('Ni3', 'L(6946)/L(7890)', 'RMS([E(6946), E(7890)])')
diags_dict['[NiIII] 6000/8500'] = ('Ni3', 'L(6000)/L(8500)', 'RMS([E(6000), E(8500)])')
diags_dict['[NiIII] 6401/8500'] = ('Ni3', 'L(6401)/L(8500)', 'RMS([E(6401), E(8500)])')
diags_dict['[NiIII] 6534/8500'] = ('Ni3', 'L(6534)/L(8500)', 'RMS([E(6534), E(8500)])')
diags_dict['[NiIII] 6682/8500'] = ('Ni3', 'L(6682)/L(8500)', 'RMS([E(6682), E(8500)])')
diags_dict['[NiIII] 6797/8500'] = ('Ni3', 'L(6797)/L(8500)', 'RMS([E(6797), E(8500)])')
diags_dict['[NiIII] 6946/8500'] = ('Ni3', 'L(6946)/L(8500)', 'RMS([E(6946), E(8500)])')
diags_dict['[NiIII] 6000/7125'] = ('Ni3', 'L(6000)/L(7125)', 'RMS([E(6000), E(7125)])')
diags_dict['[NiIII] 6401/7125'] = ('Ni3', 'L(6401)/L(7125)', 'RMS([E(6401), E(7125)])')
diags_dict['[NiIII] 6534/7125'] = ('Ni3', 'L(6534)/L(7125)', 'RMS([E(6534), E(7125)])')
diags_dict['[NiIII] 6682/7125'] = ('Ni3', 'L(6682)/L(7125)', 'RMS([E(6682), E(7125)])')
diags_dict['[NiIII] 6797/7125'] = ('Ni3', 'L(6797)/L(7125)', 'RMS([E(6797), E(7125)])')
diags_dict['[NiIII] 6946/7125'] = ('Ni3', 'L(6946)/L(7125)', 'RMS([E(6946), E(7125)])')
diags_dict['[NiIII] 6000/6401'] = ('Ni3', 'L(6000)/L(6401)', 'RMS([E(6000), E(6401)])')
diags_dict['[NiIII] 6534/6401'] = ('Ni3', 'L(6534)/L(6401)', 'RMS([E(6534), E(6401)])')
diags_dict['[NiIII] 6946/6401'] = ('Ni3', 'L(6946)/L(6401)', 'RMS([E(6946), E(6401)])')
diags_dict['[NiIII] 6000/6797'] = ('Ni3', 'L(6000)/L(6797)', 'RMS([E(6000), E(6797)])')
diags_dict['[NiIII] 6534/6797'] = ('Ni3', 'L(6534)/L(6797)', 'RMS([E(6534), E(6797)])')
diags_dict['[NiIII] 6946/6797'] = ('Ni3', 'L(6946)/L(6797)', 'RMS([E(6946), E(6797)])')
diags_dict['[NiIII] 6000/6682'] = ('Ni3', 'L(6000)/L(6682)', 'RMS([E(6000), E(6682)])')
diags_dict['[NiIII] 6534/6682'] = ('Ni3', 'L(6534)/L(6682)', 'RMS([E(6534), E(6682)])')
diags_dict['[NiIII] 6946/6682'] = ('Ni3', 'L(6946)/L(6682)', 'RMS([E(6946), E(6682)])')

class Diagnostics(object):
    """
    Diagnostics is the class used to manage the diagnostics and to computed physical conditions 
        (electron temperatures and densities) from them. 
    It is also the class that plots the diagnostic Te-Ne diagrams.

    """    
    def __init__(self, addAll=False, OmegaInterp='Cheb', NLevels=None):
        """
        Diagnostics constructor
        
        Parameters:
            - addAll        switch to include all defined diagnostics (default = False)
            - OmegaInterp   parameter sent to Atom, default is 'Cheb', other can be 'linear'
            
        """
        self.log_ = pn.log_ 
        self.calling = 'Diagnostics'
        ##            
        # @var diags
        # The dictionary containing the diagnostics
        self.diags = {}
        ##
        # @var atomDict
        # The dictionary containing the atoms used for the diagnostics        
        self.atomDict = {}
        self.OmegaInterp = OmegaInterp
        self.NLevels = NLevels
        if addAll:
            self.addAll()
        self.ANN_n_tem=30
        self.ANN_n_den=30
        self.ANN_inst_kwargs = {'RM_type' : 'SK_ANN', 
                                'verbose' : False, 
                                'scaling' : True,
                                'use_log' : True
                                }
        self.ANN_init_kwargs = {'solver' : 'lbfgs', 
                                'activation' : 'tanh', 
                                'hidden_layer_sizes' : (10, 30, 10), 
                                'max_iter' : 20000
                                }


    def getDiagFromLabel(self, label):
        """
        Return the definition of a diagnostic (the 3 or 4 elements tuple)
        
        Usage:
            diags.getDiagFromLabel('[NII] 5755/6548')
        
        Parameters:
            -label a diagnostic label 
            
        """
        if label in self.diags:
            return self.diags[label]
        else:
            self.log_.warn('{0} not in diagnostic'.format(label), calling=self.calling)
            return None

        
    def getDiags(self):
        """
        Return the definitions (tuples) of all the diagnostics defined in the Object.
        No parameter.
        
        """
        return [self.getDiagFromLabel(label) for label in self.diags]

            
    def getDiagLabels(self):
        """
        Return the labels of all the diagnostics defined in the Object
        No parameter.
        
        """
        return self.diags.keys()    

    
    def getAllDiags(self):
        """
        Return the definitions (tuples) of all the possible diagnostics.
        No parameter.
        
        """
        return diags_dict

            
    def getAllDiagLabels(self):
        """
        Return the labels of all the possible diagnostics.
        No parameter.
        
        """
        return diags_dict.keys()    

    
    def getUniqueAtoms(self):
        """
        Return a numpy.ndarray of the ions needed by the diagnostics. Unique is applied to the list before returning.
        No parameter.
        
        """
        return np.unique([self.diags[d][0] for d in self.diags if self.diags[d] is not None])
    
    
    def addDiag(self, label=None, diag_tuple=None, atom=None, wave_range=None):
        """
        Add diagnostics to the list of available diagnostics. It can either be one of the built-in diagnostics,
        a new, user-defined one, or a subset of the built-in diagnostics corresponding to a given atom or wavelength.
        
        Parameters:
            - label        a string or a list of strings describing the diagnostic, e.g. '[OIII] 4363/5007'. 
                           If it is not a key of diags_dict (a diagnostic define by PyNeb), diag_tuple must also be specified
            - diag_tuple   a 3 elements tuple containing:
                           + the atom, e.g. 'Ar5'
                           + the algebraic description of the diagnostic, in terms of line wavelengths or blends or levels, 
                             e.g. '(L(6435)+L(7006))/L(4626)'
                           + the algebraic description of the error, e.g. 
                             'RMS([E(6435)*L(6435)/(L(6435)+L(7006)), E(7006)*L(7006)/(L(6435)+L(7006)), E(4626)])'
            - atom         the selected atom, e.g. 'O3'
            - wave_range   the selected wavelength range
            
        Usage:
        diags.addDiag('[OIII] 4363/5007')
        diags.addDiag('[OIII] 5007/51m', ('O3', 'L(5007)/L(51800)', 'RMS([E(51800), E(5007)])'))
        diags.addDiag(atom='O3')
        diags.addDiag(wave_range=[4000, 6000])

        """
        if type(label) is list:
            for lab in label:
                self.addDiag(lab)
        elif label in self.diags:
            self.log_.warn('{0} already in diagnostic'.format(label), calling=self.calling)
        elif label in diags_dict:
            self.log_.message('Adding diag {0}'.format(label), calling=self.calling)
            self.diags[label] = diags_dict[label]
            atom = diags_dict[label][0]
        elif type(diag_tuple) is tuple:
            if len(diag_tuple) == 3:    
                self.diags[label] = diag_tuple
                atom = diag_tuple[0]
            else:
                self.log_.error('{0} is not in the list of diagnostics. The parameter diag_tuple must be a 3-elements tuple describing the diagnostic'.format(label), calling=self.calling + '.addDiag')
        elif atom is not None:
            for label in diags_dict:
                if diags_dict[label][0] == atom:
                    self.addDiag(label)
#                    if atom not in self.atomDict:
#                        self.atomDict[atom] = pn.Atom(parseAtom(atom)[0], parseAtom(atom)[1])
        elif wave_range is not None:
# To be done: add the possibility to use several independent ranges. The following bit does the trick, but
# only if all the wavelengths of a diagnostic are included in the same range
#            if type(wave_range[0]) is list:
#                for subrange in wave_range:
#                    self.addDiag(label, diag_tuple, atom, subrange)
                 
            for label in diags_dict:
                expr = diags_dict[label][1]
                waves = []
                wave = ''
                in_wave = False
                for i in expr:
                    if i.isdigit():
                        wave = wave + i
                        in_wave = True
                    elif (i == 'm'):
                        if in_wave == True:
                            waves.append(int(wave) * 10000.)
                            in_wave = False
                            wave = ''
                    else:
                        if in_wave == True:
                            waves.append(int(wave))
                            in_wave = False
                            wave = ''
                if (min(waves) > min(wave_range)) and (max(waves) < max(wave_range)):
                    self.addDiag(label)
        else:
            self.log_.error('Bad syntax. You must either give the label of an existing diagnostic, or the label and the tuple of a new one,' + 
                            'or an ion, or a wave range. label={0}, diag_tuple={1}, atom={2}, wave_range={3}'.format(label, diag_tuple, atom, wave_range),
                            calling=self.calling + '.addDiag')
        if atom not in self.atomDict and (type(label) is not list):
            self.atomDict[atom] = pn.Atom(parseAtom(atom)[0], parseAtom(atom)[1], NLevels=self.NLevels)
               
               
    def addAll(self):
        """
        Insert all the possible diagnostics in the Object (no parameters)
        
        """
        for label in diags_dict:
            self.addDiag(label)
            
            
    def delDiag(self, label=None):
        """
        Remove a diagnostic, based on its label.
        
        Parameter:
            - label  the label of the diagnostic ('all': removes all the diagnostics)

        """
        if label == 'all':
            for item in self.diags:
                self.delDiag(item)
        elif label in self.diags:
            del self.diags[label]
        else:
            self.log_.warn('{0} not in diagnostic, cannot be removed'.format(label), calling=self.calling)
    
    
    def addDiagsFromObs(self, obs):
        """
        Add all the possible diagnostics that can be computed from an Observation object
        
        Usage:
            diags.addDiagsFromObs(obs)
            
        Parameter:
            obs     an Observation object
        """
        
        if not isinstance(obs, pn.Observation):
            pn.log_.error('The argument must be an Observation object', calling=self.calling + 'addDiagsFromObs')
        old_level = pn.log_.level
        def I(i, j):
            wave = atom.wave_Ang[i - 1, j - 1]
            corrIntens = obs.getLine(sym, spec, wave).corrIntens
            return corrIntens
        def L(wave):
            corrIntens = obs.getLine(sym, spec, wave).corrIntens
            return corrIntens
        def B(label):
            full_label = atom + '_' + label
            corrIntens = obs.getLine(label=full_label).corrIntens
            return corrIntens
        def S(label):
            full_label = atom + '_' + label + 'A'
            corrIntens = obs.getLine(label=full_label).corrIntens
            return corrIntens
            
        for label in diags_dict:
            atom, diag_expression, error = diags_dict[label]
            sym, spec, rec = parseAtom2(atom)
            if label == '[OII] 3727+/7325+c':
                try:
                    diag_value = eval(diag_expression)
                except Exception as ex:
                    pn.log_.level = old_level
                    pn.log_.debug('Diag not valid {} {}'.format(label, diag_expression))
            try:
                pn.log_.level = 1
                diag_value = eval(diag_expression)
                pn.log_.level = old_level
                if atom not in self.atomDict:
                    if rec == 'r':
                        self.atomDict[atom] = pn.RecAtom(atom=sym+spec)
                    else:
                        self.atomDict[atom] = pn.Atom(atom=atom, NLevels=self.NLevels)
                self.addDiag(label)
            except Exception as ex:
                pn.log_.level = old_level
                pn.log_.debug('Diag not valid {} {}'.format(label, diag_expression))
                
    
    def setAtoms(self, atom_dic):
        """
        Define the dictionary containing the atoms used for the diagnostics.
        A dictionary of atom instantiations refereed by atom strings, for example:
        {'O3': pn.Atom('O', 3)}
        
        """
        if type(atom_dic) != type({}):
            self.log_.error('the parameter must be a dictionary.', calling=self.calling + '.setAtoms')
            return None
        for atom in atom_dic:
            if not isinstance(atom_dic[atom], pn.Atom) and not isinstance(atom_dic[atom], pn.RecAtom):
                self.log_.error('the parameter must be a dictionary of Atom.', calling=self.calling + '.setAtoms')
                return None
        self.atomDict = atom_dic
    
    
    def addClabel(self, label, clabel):
        """
        Add an alternative label to a diagnostic that can be used when plotting diagnostic diagrams.
        
        """
        if label in self.diags:
            self.diags[label] = (self.diags[label][0], self.diags[label][1], self.diags[label][2], clabel)
        else:
            pn.log_.warn('Try to add clabel in undefined label {0}'.format(label), calling=self.calling)
    
    
    def plot(self, emis_grids, obs, quad=True, i_obs=None, alpha=0.3, ax=None, error_band=True):
        """
        PLotting tool to generate Te-Ne diagrams.
        
        Usage:
            diags.plot(emisgrids, obs, i_obs=3)
            
        Parameters:
            - emis_grids    A dictionary of EmisGrid objects refereed by their atom strings (e.g. 'O3')
                            This can for example be the output of pn.getEmisGridDict()
            - obs           A pn.Observation object that is supposed to contain the line intensities
                            used for the plot (corrected intensities).
            - quad          If True (default) the sum of the error is quadratic,otherwise is linear.
            - i_obs         reference for the observation to be plotted, in case there is more than one
                            in the obs object
            - alpha         Transparency for the error bands in the plot
            - error_band    Boolean: plot [default] an error band
            
        """
        if not pn.config.INSTALLED['plt']: 
            pn.log_.error('Matplotlib not available, no plot', calling=self.calling + '.plot')
            return None
        else:
            import matplotlib.pyplot as plt
        if type(emis_grids) != type({}):
            self.log_.error('the first parameter must be a dictionary', calling=self.calling + '.plot')
            return None
        for em in emis_grids:
            if not isinstance(emis_grids[em], pn.EmisGrid):
                self.log_.error('the first parameter must be a dictionary of EmisGrid', calling=self.calling + '.plot')
                return None
        if not isinstance(obs, pn.Observation):
            self.log_.error('the second parameter must be an Observation', calling=self.calling + '.plot')
            return None
        if (i_obs is None) and (obs.n_obs != 1):
            self.log_.error('i_obs must be specified when obs is multiple. try i_obs=0', calling=self.calling)
            return None
        if ax is None:
            f, ax = plt.subplots()
        else:
            f = plt.gcf()
        X = np.log10(emis_grids[list(emis_grids.keys())[0]].den2D)
        Y = emis_grids[list(emis_grids.keys())[0]].tem2D
        for label in self.diags:
            self.log_.message('plotting {0}'.format(label), calling=self.calling)
            diag = self.diags[label]
            atom = diag[0]
            def I(i, j):
                return emis_grids[atom].getGrid(lev_i=i, lev_j=j)
            def L(wave):
                return emis_grids[atom].getGrid(wave=wave)
            def S(label):
                return emis_grids[atom].getGrid(label=label)
            def B(label, I=I, L=L):
                full_label = atom + '_' + label
                if full_label in BLEND_LIST:
                    to_eval = BLEND_LIST[full_label]
                else:
                    self.log_.warn('{0} not in BLEND_LIST'.format(full_label), calling=self.calling)
                    return None
                return eval(to_eval)
            diag_map = eval(diag[1])
            try:
                diag_map = eval(diag[1])
            except:
                diag_map = None
                self.log_.warn('diag {0} {1} not used'.format(diag[0], diag[1]), calling=self.calling)
            if diag_map is not None:
                sym, spec, rec = parseAtom2(atom)
                def I(i, j):
                    wave = emis_grids[atom].atom.wave_Ang[i - 1, j - 1]
                    corrIntens = obs.getLine(sym, spec, wave).corrIntens
                    if i_obs is None:
                        return corrIntens
                    else:
                        return corrIntens[i_obs]
                def L(wave):
                    corrIntens = obs.getLine(sym, spec, wave).corrIntens
                    if i_obs is None:
                        return corrIntens
                    else:
                        return corrIntens[i_obs]
                def S(label):
                    full_label = atom + '_' + label + 'A'
                    corrIntens = obs.getLine(label=full_label).corrIntens
                    if i_obs is None:
                        return corrIntens
                    else:
                        return corrIntens[i_obs]                    
                def B(label):
                    full_label = atom + '_' + label
                    corrIntens = obs.getLine(label=full_label).corrIntens
                    if i_obs is None:
                        return corrIntens
                    else:
                        return corrIntens[i_obs]
                try:
                    ee = eval(diag[1])
                    if type(ee) == np.float64:
                        diag_value = ee
                    else:
                        diag_value = ee[0]
                except:
                    #print(ee, type(ee))
                    diag_value = None
                    self.log_.warn('A line in diagnostic {0} of {1}{2} is missing'.format(diag[1], sym, spec),
                                   calling=self.calling)
                if diag_value is not None and not np.isnan(diag_value) and not np.isinf(diag_value):
                    def E(wave):
                        err = obs.getLine(sym, spec, wave).corrError
                        if i_obs is None:
                            return err
                        else:
                            return err[i_obs]
                    def BE(label):
                        full_label = atom + '_' + label
                        err = obs.getLine(label=full_label).corrError
                        if i_obs is None:
                            return err
                        else:
                            return err[i_obs]
                    def SE(label):
                        full_label = atom + '_' + label + 'A'
                        err = obs.getLine(label=full_label).corrError
                        if i_obs is None:
                            return err
                        else:
                            return err[i_obs]                        
                    if quad is True:
                        RMS = lambda err: np.sqrt((np.asarray(err) ** 2.).sum())
                    else:
                        RMS = lambda err: (np.asarray(err)).sum() 
                    tol_value = eval(diag[2])
                    col_dic = {'C':'cyan', 'N':'blue', 'O':'green', 'Ne':'magenta',
                               'Ar':'red', 'Cl':'magenta', 'S':'black', 'Fe':'blue'}
                    if sym in col_dic:
                        col = col_dic[sym]
                    else:
                        col = 'black'
                    style_dic = {'1':'-', '2':'--', '3':':', '4':'-.', '5':'-', '6':'--'}
                    style = style_dic[spec]
                    if tol_value > 0. and error_band:
                        levels = [(1 - tol_value) * diag_value, (1 + tol_value) * diag_value]
                        if levels[0] < levels[1]:
                            #pn.log_.debug('{} levels {}'.format(label, levels), calling=self.calling)
                            CS = ax.contourf(X, Y, diag_map, levels=levels, alpha=alpha, colors=col)
                    CS = ax.contour(X, Y, diag_map, levels=[diag_value], colors=col, linestyles=style)
                    try:
                        ax.set_xlabel(r'log(n$_{\rm e}$) [cm$^{-3}$]')
                        ax.set_ylabel(r'T$_{\rm e}$ [K]')
                        if len(diag) >= 4:
                            fmt = diag[3]
                        else:
                            fmt = '[{0}{1}]'.format(sym, int_to_roman(int(spec)))
                        ax.clabel(CS, inline=True, fmt=fmt, fontsize=15, colors=col)
                        if type(diag_value) is np.ndarray:
                            diag_value = diag_value[0]
                        self.log_.message('plotted {0}: {1} = {2:.2} with error of {3:.2} %'.format(fmt, label, diag_value, tol_value * 100),
                                          calling=self.calling)
                    except:
                        self.log_.message('NOT plotted {0} {1}'.format(fmt, label),
                                          calling=self.calling)
                        
        
        
    def getCrossTemDen(self, diag_tem, diag_den, value_tem=None, value_den=None, obs=None, i_obs=None,
                       guess_tem=10000, tol_tem=1., tol_den=1., max_iter=5, maxError=1e-3,
                       start_tem= -1, end_tem= -1, start_den= -1, end_den= -1,use_ANN=False, limit_res=False):
        """
        Cross-converge the temperature and density derived from two sensitive line ratios, by inputting the quantity 
        derived with one line ratio into the other and then iterating.
        The temperature- and density-sensitive ratios can be input directly or as an Observation object
    
        Parameters:
    
        - diag_tem   temperature-sensitive diagnostic line ratio
        - diag_den   density-sensitive diagnostic line ratio
        - value_tem  value of the temperature-sensitive diagnostic
        - value_den  value of the density-sensitive diagnostic
        - obs        np.Observation object. Values for observed temperature and density diagnostics are
                        taken from it if value_tem and value_den are None
        - i_obs      index of the observations to be used from obs. 
                        If None, all the observations are considered.
                        May be a boolean array
        - guess_tem  temperature assumed in the first iteration, in K
        - tol_tem    tolerance of the temperature result, in %
        - tol_den    tolerance of the density result, in %
        - max_iter   maximum number of iterations to be performed, after which the function will throw a result
        - maxError   maximum error in the calls to getTemDen, in %
        - start_tem, end_tem  lower and upper limit of the explored temperature range 
        - start_den, end_den  lower and upper limit of the explored density range 
        - use_ANN    if True, an Analogic Neural Network will be used for the determination of Te and Ne.
                        manage_RM from mwinai library is used.
                        the hyper_parameters can be set-up with the self.ANN_inst_kwargs and
                        self.ANN_init_kwargs dictionnaries.
                        self.ANN_n_tem=30 and self.ANN_n_den=30 are the number of Te and Ne
                        used to train. May also be changed before calling getCrossTemDen
        - limit_res  in case of using ANN, if limit_res, the tem and den values out of the start_tem, end_tem,
                     start_den, end_den are set to np.nan. Otherwise, extrapolation is allowed.

    
        Example:
            tem, den = diags.getCrossTemDen('[OIII] 4363/5007', '[SII] 6731/6716', 0.0050, 1.0, 
                        guess_tem=10000, tol_tem = 1., tol_den = 1., max_iter = 5)

        """
        # TODO:
        # Define a lower/upper density and temperature 
        # to be used in case the diag is pointing to values out of the boundaires
        if diag_tem not in self.diags:
            self.addDiag(diag_tem)
        atom_tem = self.diags[diag_tem][0]
        elem_tem, spec_tem = parseAtom(atom_tem)
        if atom_tem not in self.atomDict:
            self.atomDict[atom_tem] = pn.Atom(elem_tem, spec_tem, self.OmegaInterp, NLevels=self.NLevels)
        atom_tem = self.atomDict[atom_tem]
        if diag_den not in self.diags:
            self.addDiag(diag_den)
        atom_den = self.diags[diag_den][0]
        elem_den, spec_den = parseAtom(self.diags[diag_den][0])
        if (atom_den) not in self.atomDict:
            self.atomDict[atom_den] = pn.Atom(elem_den, spec_den, self.OmegaInterp, NLevels=self.NLevels)
        atom_den = self.atomDict[atom_den]
        eval_tem = self.diags[diag_tem][1]
        eval_den = self.diags[diag_den][1]
        calling = 'Diag.getCrossTemDen %s %s' % (diag_tem, diag_den)
        if value_tem is None:
            def L(wave):
                if i_obs is None:
                    return obs.getLine(elem_tem, spec_tem, wave).corrIntens
                else:
                    return obs.getLine(elem_tem, spec_tem, wave).corrIntens[i_obs]
            def I(i, j):
                wave = atom_tem.wave_Ang[i - 1, j - 1]
                if i_obs is None:
                    return obs.getLine(elem_tem, spec_tem, wave).corrIntens
                else:
                    return obs.getLine(elem_tem, spec_tem, wave).corrIntens[i_obs]
            def B(label, I=I, L=L):
                full_label = elem_tem + spec_tem + '_' + label
                if i_obs is None:
                    corrIntens = obs.getLine(label=full_label).corrIntens
                else:
                    corrIntens = obs.getLine(label=full_label).corrIntens[i_obs]
                return corrIntens
            #pn.log_.debug('to eval is {0}'.format(eval_tem), calling=calling + 'TEST')
            try:
                value_tem = eval(eval_tem)
                #pn.log_.debug('to eval = {0}'.format(value_tem), calling=calling + 'TEST')
            except:
                pn.log_.warn('No value for {0} {1}: {2} in obs'.format(elem_tem, spec_tem, diag_tem), calling=calling)
                return None
        else:
            if type(value_tem) == type([]): value_tem = np.asarray(value_tem)
        if value_den is None:
            def L(wave): 
                if i_obs is None:
                    return obs.getLine(elem_den, spec_den, wave).corrIntens
                else:
                    return obs.getLine(elem_den, spec_den, wave).corrIntens[i_obs]
            def I(i, j):
                wave = atom_den.wave_Ang[i - 1, j - 1]
                pn.log_.debug('wave is {0}'.format(wave), calling=calling + 'TEST3')
                if i_obs is None:
                    return obs.getLine(elem_den, spec_den, wave).corrIntens
                else:
                    return obs.getLine(elem_den, spec_den, wave).corrIntens[i_obs]
            def B(label, I=I, L=L):
                full_label = elem_den + spec_den + '_' + label
                if i_obs is None:
                    corrIntens = obs.getLine(label=full_label).corrIntens
                else:
                    corrIntens = obs.getLine(label=full_label).corrIntens[i_obs]
                return corrIntens
            #pn.log_.debug('to eval is {0}'.format(eval_den), calling=calling + ' TEST')
            try:
                value_den = eval(eval_den)
                #pn.log_.debug('to eval = {0}'.format(value_den), calling=calling + ' TEST1')
            except:
                pn.log_.warn('No value for {0} {1}: {2} in obs'.format(elem_den, spec_den, diag_den), calling=calling)
                return None
        else:
            if type(value_den) == type([]): value_den = np.asarray(value_den)
        if use_ANN:
            try:
                from ai4neb import manage_RM
                ai4neb_OK = True
            except:
                ai4neb_OK = False
                pn.log_.error('ai4neb is not installed')
            if ai4neb_OK:      
                if start_tem == -1:
                    tem_min = 3000.
                else:
                    tem_min = start_tem
                if end_tem == -1:
                    tem_max = 20000.
                else:
                    tem_max = end_tem
                if start_den == -1:
                    den_min = 10.
                else:
                    den_min = start_den
                if end_den == -1:
                    den_max = 1e6
                else:
                    den_max = end_den
                # define emisGrid objects to generate Te-Ne emissionmaps
                tem_EG = pn.EmisGrid(atomObj=atom_tem, 
                                     n_tem=self.ANN_n_tem, n_den=self.ANN_n_den, 
                                     tem_min=tem_min, tem_max=tem_max,
                                     den_min=den_min, den_max=den_max)
                den_EG = pn.EmisGrid(atomObj=atom_den, 
                                     n_tem=self.ANN_n_tem, n_den=self.ANN_n_den, 
                                     tem_min=tem_min, tem_max=tem_max,
                                     den_min=den_min, den_max=den_max)
                # compute emission line ratio maps in the Te-Ne space
                tem_2D = tem_EG.getGrid(to_eval = eval_tem)
                den_2D = den_EG.getGrid(to_eval = eval_den)
                # X is a set of line ratio pairs
                X = np.array((tem_2D.ravel(), den_2D.ravel())).T
                # y is a set of corresponding Te-Ne pairs
                y = np.array((tem_EG.tem2D.ravel()/1e4, np.log10(den_EG.den2D.ravel()))).T
                # Instantiate, init and train the ANN
                self.ANN = manage_RM(X_train=X, y_train=y, **self.ANN_inst_kwargs)
                self.ANN.init_RM(**self.ANN_init_kwargs)
                self.ANN.train_RM()
                # set the test values to the one we are looking for
                self.ANN.set_test(np.array((value_tem, value_den)).T)
                # predict the result and denormalize them
                self.ANN.predict()
                if self.ANN.isfin is None:
                    tem = self.ANN.pred[:,0]*1e4
                    den = 10**self.ANN.pred[:,1]
                else:
                    tem = np.zeros_like(value_tem) * -10
                    tem[self.ANN.isfin] = self.ANN.pred[:,0]*1e4
                    den = np.zeros_like(value_tem) * -10
                    den[self.ANN.isfin] = 10**self.ANN.pred[:,1]
                if limit_res:
                    tem[(tem<tem_min) | (tem>tem_max)] = np.nan
                    den[(den<den_min) | (den>den_max)] = np.nan
        else:
            den = atom_den.getTemDen(value_den, tem=guess_tem, to_eval=eval_den,
                                     maxError=maxError, start_x=start_den, end_x=end_den)
            
            tem = atom_tem.getTemDen(value_tem, den=den, to_eval=eval_tem,
                                     maxError=maxError, start_x=start_tem, end_x=end_tem)
    #        self.log_.debug('tem: ' + str(tem) + ' den:' + str(den), calling='getCrossTemDen')
            no_conv = np.ones_like(den).astype(bool)
            n_tot = np.asarray(value_tem).size
            for i in np.arange(max_iter):
                if type(tem) == type(1.):
                    tem_old = tem
                else:
                    tem_old = tem.copy()
                if type(den) == type(1.):
                    den_old = den
                else:
                    den_old = den.copy()
                
                if n_tot > 1:
                    den[no_conv] = atom_den.getTemDen(value_den[no_conv], tem=tem_old[no_conv],
                                                      to_eval=eval_den, start_x=start_den, end_x=end_den)
                    tem[no_conv] = atom_tem.getTemDen(value_tem[no_conv], den=den_old[no_conv],
                                                      to_eval=eval_tem, start_x=start_tem, end_x=end_tem)
                else:
                    den = atom_den.getTemDen(value_den, tem=tem_old, to_eval=eval_den, start_x=start_den, end_x=end_den)
                    tem = atom_tem.getTemDen(value_tem, den=den_old, to_eval=eval_tem, start_x=start_tem, end_x=end_tem)
                    
                no_conv = ((abs(den_old - den) / den * 100) > tol_den) | ((abs(tem_old - tem) / tem * 100) > tol_tem)
                if type(no_conv) == type(True):
                    n_no_conv = int(no_conv)
                else:
                    n_no_conv = no_conv.sum()
                
                pn.log_.message('{0} (max={1}): not converged {2} of {3}.'.format(i, max_iter, n_no_conv, n_tot),
                                calling=calling)
                if n_no_conv == 0:
                    return tem, den
            if n_tot == 1:
                tem = np.nan
                den = np.nan
            else:
                tem[no_conv] = np.nan
                den[no_conv] = np.nan
        return tem, den
    
    def getDiagLimits(self, diag):
        """
        Return the low and high density values for a given diagnostic
        """
        atom = self.atomDict[self.diags[diag][0]]
        to_eval = self.diags[diag][1]
        HDR = atom.getHighDensRatio(to_eval = to_eval)
        LDR = atom.getLowDensRatio(to_eval = to_eval)
        return(np.sort((LDR, HDR)))
    
    


"""@module plotAtomicData
Module to plot the atomic data
""" 

import pyneb as pn
import numpy as np
if pn.config.INSTALLED['plt']:
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MaxNLocator

from pyneb.utils.misc import int_to_roman, parseAtom

class DataPlot(object):
    """
    Plot transition probabilities and collision strengths from different data sets
    
    """
    def __init__(self, elem=None, spec=None, all_data=[], atom=None, n_tem_points=10000, 
                 ref_tem=None, OmegaInterp='Cheb',NLevels=None):
        """
    Parameters:
        - elem         atomic elem 
        - spec       ionization stage in spectroscopic notation (I = 1, II = 2, etc.)
        - atom        e.g. 'O3'
        - all_data       dictionary of all_data to be compared (see above for format)
        - [n_tem_points] number of points in the fit (default=100; increase if fit is not smooth)
        - [ref_tem]      array of temperature values to be signaled in the plots
        - OmegaInterp    interpolating function between Omega values ('Cheb' [default], 'Linear')
    
    Example:
        dataplot = pn.DataPlot('O', 3) # initializes the plot
        dataplot.plotA() # transition probabilities plot 
        dataplot.plotRelA() # relative transition probabilities plot
        dataplot.plotOmega() # collision strength plot    

        
        """
        colors = np.array(['r', 'g', 'b', 'm', 'c', 'y'])
        self.calling = 'DataPlot'
        if atom is not None:
            self.atom = str.capitalize(atom)
            self.elem = parseAtom(self.atom)[0]
            self.spec = int(parseAtom(self.atom)[1])
        else:
            self.elem = str.capitalize(elem)
            self.spec = int(spec)
            self.atom = self.elem + str(self.spec)

        old_data = pn.atomicData.getDataFile(self.atom)
        # Check if matplotlib installed
        if not pn.config.INSTALLED['plt']:
            pn.log_.warn('Matplotlib not installed!', calling=self.calling)

        # Separate omega and A data sets    
        atom_data = []
        coll_data = []
        if all_data == []:
            all_data = pn.atomicData.getAllAvailableFiles(self.atom)
        i_colors = 0    
        for file_ in all_data:
            ID = file_.split('_')[-1]
            type_ = (file_.split('_')[2]).split('.')[0]
            if type_ in ['atom', 'coll']:
                data_list = type_ + '_data'
                vars()[data_list].append({'ID': ID, 'file_': file_, 'type': type_, 'color': colors[i_colors % len(colors)]})
                i_colors += 1
                
        self.atom_data = atom_data
        self.coll_data = coll_data

        # Temperature values to be signaled by vertical lines in omega plots. Can be changed by defining ref_tem
        if ref_tem is None:
            self.ref_tem = np.log10([5000., 10000., 20000.]) 
        else:
            self.ref_tem = ref_tem

        # If it's not standard effective collision strengths, we don't want it
        coll_data_temp=[]
        for item in self.coll_data: 
            coll_data_temp.append(item)
        for data in coll_data_temp:
            pn.atomicData.setDataFile(data['file_'])
            atom = pn.Atom(self.elem, self.spec, OmegaInterp=OmegaInterp, NLevels=NLevels)
            """
            try:
                if 'O_UNIT' in atom.CollData.comments.keys():
                    self.coll_data.remove(data)
            except:
                pass
            """
        # For each data set, an atom is built. 
        del(atom)
        for data in self.atom_data + self.coll_data:
            pn.atomicData.setDataFile(data['file_'])
            atom = pn.Atom(self.elem, self.spec, OmegaInterp=OmegaInterp, NLevels=NLevels)
            data['atom'] = atom
        for data in old_data:
            if data is not None:
                pn.atomicData.setDataFile(data)
        self.atom_rom = (self.elem + '_' + int_to_roman(int(self.spec))).lower()
        self.n_tem_points = n_tem_points

        self.atom_n_max = 0
        for at in self.atom_data:
            if at['atom'].atomFileType != 'chianti':
                if at['atom'].atomNLevels > self.atom_n_max:
                    self.atom_n_max = at['atom'].atomNLevels
        self.coll_n_max = 0
        for at in self.coll_data:
            if at['atom'].collFileType != 'chianti':
                if at['atom'].collNLevels > self.coll_n_max:
                    self.coll_n_max = at['atom'].collNLevels
                                    
    def plotA(self, save=False, figsize=(18, 12), fignum=None, NLevels=None):
        """
        Plot the log of the A values of each data set 
        
        Parameters:
            - save     if True, saves the plot in a file
            - figsize  figure size (default: [18, 12])
            - fignum    figure Number

        """
        if NLevels is None:
            atom_n_max = self.atom_n_max
        else:
            atom_n_max = NLevels
        if not pn.config.INSTALLED['plt']:
            pn.log_.error('Matplotlib not installed!', calling=self.calling)
        plt.figure(fignum, figsize=figsize)
        plt.clf()
        ax = plt.subplot(111)
        x = np.arange(atom_n_max * (atom_n_max - 1) / 2.)
        ticks = x
        # Inventory of markers. Chosen to be distinguished even when overlapped
        mark = ['<', (5, 1), '>', 'o', '|']
        i_marker = 0
        # Background colors to distinguish lower levels 
        bg_color = 'grbymc' 
        tick_label = []
        
        for i in range(atom_n_max - 1):
            # x0, x1 are start and end points of each level
            x0 = i * (atom_n_max - 1) - i * (i - 1) / 2 - 0.5
            width = atom_n_max - i - 1
            x1 = x0 + width
            plt.axvspan(x0, x1, facecolor=bg_color[i%6], alpha=0.05)
            # The x axis must stretch the maximum range (although some data set might have a lower n_level)
            for j in range(i + 1, atom_n_max):
                tick_label.append('(' + str(j + 1) + ', ' + str(i + 1) + ')')


        for data in self.atom_data:
            n_levels = data['atom'].atomNLevels
            Ay = []
            A = data['atom'].getA()
            color = data['color']
            try:
                for i in range(n_levels - 1):
                    for j in range(i + 1, n_levels):
                        if A[j, i] > 0:
                            Ay.append(np.log10(A[j, i]))
                        else:
                            Ay.append(np.NaN)
                plt.scatter(x, Ay, marker=mark[i_marker], s=300., c=color, alpha=0.35, linewidths=1, label='%s' % (data['ID']))
                ax.set_xticks(ticks)
            except:
                pn.log_.warn('Problem in plotting A', calling=self.calling + '.plotA')
            i_marker += 1
        ax.set_xticklabels(tick_label)

        # Plot features
        plt.xlabel('Transition')
        plt.ylabel('Log A(j, i)')
        plt.title('Transition probabilities for [%s %s]' % (self.elem, int_to_roman(int(self.spec))))
        plt.legend(loc='lower right', markerscale=1., scatterpoints=1, borderpad=1, labelspacing=1)
        plt.show()      
        if save:
            plt.figure(figsize=[18, 12])
            plt.savefig(self.atom_rom + '_' + data['ID'] + "-" + self.ref_data + "_A.pdf")



    def plotAllA(self, save=False, figsize=(18, 12), fignum=None, NLevels=None):
        """
        Plot the log of the A values of each data set 
        
        Parameters:
            - save     if True, saves the plot in a file
            - figsize  figure size (default: [18, 12])
            - fignum    figure Number

        """
        if NLevels is None:
            atom_n_max = self.atom_n_max
        else:
            atom_n_max = NLevels
        if not pn.config.INSTALLED['plt']:
            pn.log_.error('Matplotlib not installed!', calling=self.calling)
        fig = plt.figure(fignum, figsize=figsize)
        plt.clf()
        max_y_ticks = 6
        ticks = np.arange(len(self.atom_data)+1)[1:] 
        
        A = np.zeros([len(self.atom_data), atom_n_max, atom_n_max])
        color=[]
        for i, data in enumerate(self.atom_data):
            A_tmp = data['atom'].getA()
            if A_tmp.shape[0] > atom_n_max:
                A[i, :] = A_tmp[0:atom_n_max, 0:atom_n_max]
            else:
                A[i, 0:A_tmp.shape[0], 0:A_tmp.shape[1]] = A_tmp
            color.append(data['color'])
        
        A[np.where((A<=0))] = np.NaN
        lgA = np.log10(A)
        
        x = ticks
        color = np.array(color)
        for j in range(2, atom_n_max+1):
            for i in range(1, j):
                y = lgA[:, j-1, i-1]
                ax = plt.subplot(atom_n_max - 1, atom_n_max - 1, (atom_n_max-1)*(j-2) + i)
                ax.set_xticks(x)
                plt.xlim((min(x)-0.5, max(x)+0.5))
                ax.yaxis.set_major_locator(MaxNLocator(max_y_ticks-1))
                try:
                    plt.scatter(x, y, c=color, label='_nolegend_', s=40, edgecolor='None')
                    lbl = '({0} -> {1})'.format(j,i)
                    ax.text(0.95, 0.95, lbl, fontsize=10, color="#660066", transform=ax.transAxes, ha="right", va="top")                                    
                except:
                    pn.log_.warn('Problem with plotting a subplot {} {}'.format(i,j), calling=self.calling)
                if (j==atom_n_max) & (i==1):  
                    plt.xlabel('Reference #', fontsize=8)
                    plt.ylabel('Log(A$_{ji}$)', fontsize=8)
        plt.subplots_adjust(left=0.15, right=0.95, top=0.95, bottom=0.12)
        plt.gca().get_yaxis().get_major_formatter().set_useOffset(False)
        ax.set_xticks(x)
        
        fig.text(.5, .05, "i (lower level)", fontsize=12, ha='center', color="#660066", va='top')
        fig.text(.05, .5, "j (upper level)", ha='left', va='center', fontsize=12, color="#660066", rotation=90)
        title = "Available transition probabilities for {0} {1}".format(self.elem, int_to_roman(int(self.spec)))
        fig.text(.5, .95, title, color="#191970", fontsize=14, ha='center')
        for i, data in enumerate(self.atom_data):
            x_txt = .95
            y_txt = .85 - .04*i 
            fig.text(x_txt, y_txt, "Ref. %d: %s" %(i+1, data['ID']), color=color[i], ha = 'right') 

        plt.tight_layout(pad=0.1)
        if save:
            plt.savefig(self.atom_rom + "_all_As.pdf")
    
        plt.show()      
        

    def plotRelA(self, ref_data=None, save=False, figsize=None, fignum=None, NLevels=None, fig=None):
        """
        Plot the relative difference of the A of each data set with respect to the reference one

        Parameters:
            - ref_data       reference data set for comparing transition probabilities (default=first data ID)
            - save           if True, save the plot in a file (default: False)
            - figsize        figure size (default: [18, 12])
            - fignum         figure number

        """
        if NLevels is None:
            atom_n_max = self.atom_n_max
        else:
            atom_n_max = NLevels
        if not pn.config.INSTALLED['plt']:
            pn.log_.error('Matplotlib not installed!', calling=self.calling)
        # Commented out because it produced an extra, empty frame - VL 30 Jul 2015
        #if fig is None:
        #    fig = plt.figure(fignum, figsize=figsize)
        #plt.clf()
        ticks = range(atom_n_max + 1)[1:]

        # The As of the reference data set are stored
        if ref_data is None:
            ref_data = self.atom_data[0]['ID']
        for data in self.atom_data:
            if (data['ID'] is ref_data):
                ref_A = data['atom'].getA()
                # Only non-zero values are taken into account to prevent dividing by zero
                nonzero_ref_A_indexes = np.nonzero(ref_A)

        for data in self.atom_data:
            if (data['ID'] is not ref_data):
                A = data['atom'].getA()
                try:
                    # The data to be compared might have fewer levels than the reference one. Only the ones in common are considered.
                    tmp_indexes = np.asarray(nonzero_ref_A_indexes) 
                    up_indexes = tmp_indexes[0][tmp_indexes[0] < data['atom'].atomNLevels]
                    lo_indexes = tmp_indexes[1][tmp_indexes[0] < data['atom'].atomNLevels]
                    # Indexes for which the reference data set is not zero
                    ref_indexes = (up_indexes, lo_indexes)
                    # Ratio of A/A_ref for non-zero A_ref values
                    A_ratio = A[ref_indexes] / ref_A[ref_indexes]
                    nonzero_A_ratio_indexes = np.nonzero(A_ratio)
                    rel_A = np.log10(A_ratio[nonzero_A_ratio_indexes])
                    # Plotting starts
                    #fig = plt.figure()
                    if fig is None:
                        fig = plt.figure(fignum, figsize=figsize)
                    plt.clf()
                    ax = plt.subplot(111)
                    fig.subplots_adjust(top=0.9)
                    # Physical levels = array levels + 1
                    x = np.asarray(lo_indexes) + 1
                    y = np.asarray(up_indexes) + 1
                    # Size is proportional to difference between As
                    # Numerical value of size adjusted empirically, no particular meaning 
                    size = np.multiply(10000., np.abs(rel_A))
                    # A different color is assigned to data points depending on whether they are smaller or larger
                    # than the reference data.
                    # The ratio is not compared to exactly one to include only truly different values  
                    index_grt = np.where(rel_A > 0.0000)
                    color_grt = np.array(['#FF8C00']).repeat(x[index_grt].size)
                    index_sml = np.where(rel_A <= 0.0)
                    color_sml = np.array(['#FF1480']).repeat(x[index_sml].size)
                    # The range is adjusted to include only values for which there is a difference
                    xmin = np.max([np.min(x[index_sml]), np.min(x[index_grt])])
                    xmax = np.min([np.max(x[index_sml]), np.max(x[index_sml])])
                    ax.set_xticks(ticks)
                    ax.set_xlim((xmin - 0.25, xmax + 0.25)[:])
                    # The x-axis is plotted on the top to mimic the data array in the fits file
                    for tick in ax.xaxis.get_major_ticks():
                        tick.label1On = False
                        tick.label2On = True
                    ymin = np.min([np.min(y[index_sml]), np.min(y[index_grt])])
                    ymax = np.max([np.max(y[index_sml]), np.max(y[index_sml])])
                    ax.set_yticks(ticks)
                    # The y axis increases downward to mimic the data array in the fits file
                    ax.set_ylim((ymin - 0.25, ymax + 0.25)[::-1])
                    # Fake points, just to generate the labels (I don't know how to generate a reasonable label
                    # for data points with markers of different size)
                    plt.scatter(x[index_grt][0], y[index_grt][0], marker='o', s=.0002, c=color_grt[0],
                                label='Positive (max = %.2f)' % (np.max(rel_A)))
                    plt.scatter(x[index_grt][0], y[index_grt][0], marker='o', s=.0002, c=color_sml[0],
                                label='Negative (min = %.2f)' % (np.min(rel_A)))
                    # True data; size reflects divergence from reference data, color whether positive or negative
                    plt.scatter(x[index_grt], y[index_grt], marker='o', s=size[index_grt], c=np.asarray(color_grt), alpha=1.0)
                    plt.scatter(x[index_sml], y[index_sml], marker='o', s=size[index_sml], c=np.asarray(color_sml), alpha=1.0)
                    # X-axis label 
                    ax.text(.5, 1.05, '$i$ (lower level)', transform=ax.transAxes, ha='center', va='bottom')
                    #plt.xlabel('$i$ (lower level)')
                    # Y-axis label
                    plt.ylabel('$j$ (upper level)')
                    # Legend
                    plt.legend(loc='upper right', markerscale=1000., scatterpoints=1, borderpad=1,
                               labelspacing=1, prop=dict(size=12), title=u'Log (%s / %s)' % (data['ID'], ref_data))
                    # Plot title
                    ax.text(.5, -.10, "Relative difference between $A$s data for [%s %s]" 
                                % (self.elem, int_to_roman(int(self.spec))),
                                transform=ax.transAxes, ha='center', va='bottom', color='#191970', size=12)
                except:
                    pn.log_.warn('Problem in plotting relA', calling='DatasetPlot')

                if save:
                    plt.savefig(self.atom_rom + '_' + data['ID'] + "-" + ref_data + "_relA.pdf")
    
                plt.show()      


    def plotOmega(self, save=False, figsize=(18, 12), fignum=1, scan_orders=None, NLevels=None,
                  fig=None):
        """
        Plot the tabulated collision strengths of each data set and the fit that is performed by PyNeb
        
        Parameters:
            - save     Boolean. Determine if the plot is automatically saved in a file (default: False)
            - figsize  List. figure size in inches (default: [18, 12])
            - fignum   Figure Number
            - scan_orders = None or (min_order, max_order) or (min_order, -1) to go until the max. DEPRECATED!!!

        """
        if NLevels is None:
            coll_n_max = self.coll_n_max
        else:
            coll_n_max = NLevels
        # Plotting range somewhat larger than actual data range (less tight) 
        if not pn.config.INSTALLED['plt']:
            pn.log_.error('Matplotlib not installed!', calling=self.calling)
        if fig is None:
            fig = plt.figure(fignum, figsize=figsize)
        plt.autoscale(tight=False)
        # I need two 
        first = True
        first_done = False
        plt.clf()
        
        legend_text = []
        legend_lines = []
#        style_dic = ['-', '--', '-.', ':'] * 10
        for data in self.coll_data:
            # tem_points = tabulated temperature points (different for each data set)
            tem_points = self.tem_in_K(data['atom'].tem_units, data['atom'].getTemArray())
            tem_min = min(tem_points)
            tem_max = max(tem_points)
            # tem_funct = array of temperature values for which the fit is evaluated
            tem_funct = np.linspace(tem_min, tem_max, self.n_tem_points)
            x_dots = np.log10(tem_points)
            x_lines = np.log10(tem_funct)
            if 'COEFF' in data['atom'].CollData.comments.keys():
                coeff = float(data['atom'].CollData.comments['COEFF'])
            else:
                coeff = 1.0
            # Loops over all levels
            for i in range(1, coll_n_max+1):
                for j in range(i + 1, coll_n_max+1):                
                        # N levels require an N-1 x N-1 array of plots
                        # The subplots are arranged in rows and columns according to the upper and lower levels
 
                        ax = plt.subplot(coll_n_max-1, coll_n_max-1, 
                                         (coll_n_max-1) * (j - 2) + i)
                        if (i <= data['atom'].collNLevels) and (j <= data['atom'].collNLevels):
                            y_dots = data['atom'].getOmega(tem_points, j, i)
                            try:
                                if 'O_UNIT' not in data['atom'].CollData.comments:
                                    y_dots = coeff * data['atom'].getOmegaArray(j, i)
                                    if y_dots.sum() > 0.0:
                                        ax.plot(x_dots, y_dots, color=data['color'],
                                                marker='*', linestyle='None', label='_nolegend_', markersize=10)
                                    
                                y_dots = data['atom'].getOmega(tem_points, j, i)
                                if y_dots.sum() > 0.0:
                                    ax.plot(x_dots, y_dots, color=data['color'],
                                            marker='o', linestyle='None', label='_nolegend_')
                                y_lines = data['atom'].getOmega(tem_funct, j, i)
                                if y_lines.sum() > 0.:
                                    ax.plot(x_lines, y_lines,
                                            color=data['color'], label='_nolegend_')
                            except:
                                pn.log_.warn('Problem with plotting a data set', calling=self.calling)
# Draws vertical lines at selected temperature values (only once for each subplot, hence "first") 
                        if (first and (data['atom'].collNLevels == coll_n_max)):
                            for tem in self.ref_tem:
                                plt.axvline(tem, c='blue', alpha=0.4, ls=':')
                            lbl = "$\Omega$" + "(" + str(j) + "," + str(i) + ")"
                            ax.text(0.95, 0.95, lbl, transform=ax.transAxes, ha="right", va="top")   
                            first_done = True                                 
                        # Maximum number of ticks to avoid overcrowding
                        ax.xaxis.set_major_locator(MaxNLocator(4))
                        ax.yaxis.set_major_locator(MaxNLocator(3))
            # The line type and identifier of each data set are stored for the legend
            legend_lines.append(mpl.lines.Line2D(tem_funct, tem_funct, color=data['color'], linestyle='-'))
            legend_text.append(data['ID'])
            if first_done:
                first = False
        # The label of the reference temperature values are only plotted above the last plot
        for tem in self.ref_tem:
            ax.text(tem, ax.get_ylim()[1], ' %0.0f K' % (10 ** tem), ha="left", va="bottom", 
                    rotation=65, fontsize=12, color='#FF0000', alpha=0.35)
        # Axis labels, title and legend    
        fig.text(.5, .05, "Log($T_e/K$)", fontsize=16, ha='center', va='top')
        fig.text(.05, .5, "$\Omega$", va='center', fontsize=20, rotation=90)
        fig.text(.5, .95, "[%s %s] collision strengths" % (self.elem, int_to_roman(int(self.spec))), 
                 color="#191970", fontsize=16, ha='center')
#        plt.legend(legend_lines, legend_text, loc='upper right', borderpad=1, 
#                   labelspacing=1, bbox_to_anchor=(1, 1 * coll_n_max))
        plt.legend(legend_lines, legend_text, loc='upper right', borderpad=1, 
                   labelspacing=1, bbox_to_anchor=(1, coll_n_max - 1))
        #plt.tight_layout()
        if save:
            plt.savefig(self.atom_rom + "_CS.pdf")
        
        plt.show()


    def tem_in_K(self, tem_units, tem):
        """
        Convert the temperature from the unit of the fits file into K

        Parameters:
            - tem_units     'log(K)' or 'K/1000'
            - tem           temperature
        """
        
        if (tem_units == "log(K)"):
            return np.power(10., tem)
        elif (tem_units == "K/10000"):
            return np.multiply(tem, 1.e4)
        else: #T in Kelvin in the fits file
            return tem
                
    



import re
import gzip
import os
import numpy as np
import pyneb as pn
from misc import int_to_roman 
from physics import sym2name
from manage_atomic_data import getLevelsNIST

term2_dic = {'0':'S', '1':'P', '2':'D', '3':'F', '4':'G', '5':'H', '6':'I'}

def conv_apap(str_in):
    """
    transform "1.3-01" into 1.3e-01 
    """
    a, b = re.split('\+|-',str_in)
    if '-' in str_in:
        return float(a) / 10**float(b)
    else:
        return float(a) * 10**float(b)

def read_apap(apap_file, max_levels=10):
    if max_levels is None:
        max_levels = np.Inf
        
    term1 = []
    term2 = []
    energy = []
    J = []
    if apap_file.split('.')[-1] == 'gz':
        f = gzip.open(apap_file)
    else:
        f = open(apap_file)
    foo = f.readline()
    while True:
        line = f.readline()
        if line[0:5] == "   -1":
            break
        lev = int(line[0:5])
        if lev <= max_levels:
            term1.append(line[25:26])
            term2.append(line[27:28])
            J.append(line[29:33])
            energy.append(line[34::])
    term = [t1+term2_dic[t2] for t1,t2 in zip(term1, term2)]
    
    lev_i = []
    lev_j = []
    As = []
    Omegas = []
    
    temps_str = f.readline().split()[2::]
    temp_array = np.array([conv_apap(temp) for temp in temps_str])
    N_temp = len(temp_array)
    while True:
        line = f.readline()
        if line[0:4] == "  -1":
            break
        i = int(line[1:4])
        j = int(line[4:8])
        if (i <= max_levels) and (j <= max_levels):
            lev_i.append(i)
            lev_j.append(j)
            strg5 = line[8:8+(N_temp+1)*8].split()
            As.append(conv_apap(strg5[0]))
            Omegas.append([conv_apap(omega) for omega in strg5[1::]])
    f.close()
    
    return(term, 
           np.array(energy, dtype='float'), 
           np.array(J, dtype='float'), 
           np.array(lev_i, dtype='int'), 
           np.array(lev_j, dtype='int'), 
           np.array(As, dtype='float'), 
           np.array(Omegas, dtype='float'), 
           temp_array)

def apap2ascii(apap_file, max_levels=10, source='http://www.apap-network.org/'):
    
    atom = apap_file.split('#')[1].split('.')[0]
    spectrum = int(re.findall(r'\d+', atom)[0]) + 1
    ioniz = int_to_roman(spectrum).lower()
    elem = re.findall('[a-zA-Z]+', atom)[0]
    ion_str = '{0}{1}'.format(elem.capitalize(), spectrum)
    print('Reading NIST data for {0}'.format(ion_str))
    NIST_levels = getLevelsNIST(ion_str)
    NIST_full_term = ['{0:2s}{1:.1f}'.format(l['term'][0:2],float(l['J'])) for l in  NIST_levels]
    
    term, energy, J, lev_i, lev_j, As, Omegas, temp_array = read_apap(apap_file, max_levels=max_levels)
    N_temp = len(temp_array)

    APAP_full_term = ['{0:2s}{1:.1f}'.format(t,float(j)) for t, j in zip(term, J)]
    
    N_levels = len(energy)
    stat_weight = J * 2 + 1
        
    Omega3d = np.zeros((N_levels, N_levels, N_temp))
    for i, j, Omega in zip(lev_i, lev_j, Omegas):
        try:
            i_NIST = NIST_full_term.index(APAP_full_term[i-1])
        except:
            i_NIST = None
        try:
            j_NIST = NIST_full_term.index(APAP_full_term[j-1])
        except:
            j_NIST = None
        if (i_NIST is not None) and (j_NIST is not None):
            Omega3d[i_NIST, j_NIST] = Omega
        else:
            pn.log_.warn('Levels not found in NIST: {0} {1}'.format(APAP_full_term[i-1], 
                                                                    APAP_full_term[j-1]))
            
    f_coll = open('{0}_{1}_coll_APAP14.dat'.format(elem, ioniz),'w')
    str_ = '0   0 '
    str_ += ''.join(' {0:10.4e}'.format(T) for T in temp_array)
    str_ += '\n'
    f_coll.write(str_)
    for i in 1+np.arange(N_levels):
        for j in i+1+np.arange(N_levels-i):
            str_ = '{0} {1} '.format(i, j)
            str_ += ''.join(' {0:10.8e}'.format(O) for O in Omega3d[j-1, i-1])
            str_ += '\n'
            f_coll.write(str_)
    try:
        f_coll.write("*** ATOM {0}\n".format(sym2name[elem.capitalize()]))
    except:
        pn.log_.warn('No full name for element {0}'.format(elem.capitalize()))

    f_coll.write("""*** SPECTRUM {0}
*** N_LEVELS {1}
*** T_UNIT 'K'
*** NOTE1 'Collision strengths'
*** SOURCE1 '{2}'
""".format(spectrum, N_levels, source))
    
    f_coll.close()


def make_all(dir_ = '.', max_levels=8):
    file_liste = os.listdir('.')
    for file_ in file_liste:
        if file_[-1] == 'z':
            atom = file_.split('#')[1].split('.')[0]
            if atom == 'fe':
                apap2ascii(file_, 34)
            else:
                apap2ascii(file_, max_levels)
            print('{0} done'.format(file_))


import pyneb as pn
import numpy as np
import matplotlib.pyplot as plt

"""
Usage:
pytest
pytest test_lines.py::test_atom_data


"""


verbose = False

def rel_error(calculated, expected):
    return (abs((expected - calculated) * 1.0) / (expected * 1.0))

def test_plot_data(save_pdf = False, verbose=verbose):
    """
    This function try to make the dataplot (As and Omegas) for any of the available atoms
    """

    plt.ion()
    atom_list = pn.atomicData.getAllAtoms()
    atom_list.remove('3He2')
    for atom in atom_list:
        if verbose:
            print('Doing {0}'.format(atom))
        dp = pn.DataPlot(atom=atom, OmegaInterp='Linear')
        
        dp.atom_n_max = np.min((dp.atom_n_max, 7))
        if len(dp.atom_data) > 1:
            if verbose:
                print('Plotting {0}'.format(atom))
            dp.plotAllA(save = save_pdf)
            plt.close()
          
        dp.coll_n_max = np.min((dp.coll_n_max, 7))
        if len(dp.coll_data) > 1:
            if verbose:
                print('Plotting {0}'.format(atom))
            dp.plotOmega(save = save_pdf)
            plt.close()
        
def test_diag_list():
    
    diags = ['[ArIII] 5192/7136', 
             '[ArIII] 9.0m/21.8m',
             '[ArIV] 4740/4711',
             '[CIII] 1909/1907',
             '[ClIII] 5538/5518',
             '[NII] 121m/20.5m',
             '[NII] 5755/6584',
             '[NeIII] 15.6m/36.0m',
             '[NeIII] 3343/3930+',
             '[NeV] 14.3m/24.2m',
             '[NeV] 2973/3370+',
             '[OI] 5577/6300+',
             '[OI] 63m/147m',
             '[OII] 3726/3729',
             '[OII] 3727+/7325+',
             '[OIII] 1664+/5007',
             '[OIII] 5007/88m', 
             '[OIII] 4363/5007',
             '[OIV] 1400+/25.9m',
             '[OIV] 1401/1405',
             '[SII] 4072+/6720+',
             '[SII] 4069/4076', 
             '[SII] 6731/6716',
             '[SIII] 6312/9069',
             '[SIII] 18.7m/33.5m'
             ] 
    # Build the 25 subplots
    f, axes = plt.subplots(5, 5, figsize=(20,20))
    # loop on the diagnostics
    for i, d in enumerate(diags):
        # extract from the diagnostic dictionnary 
        # the expression to be evaluated
        to_eval = pn.diags_dict[d][1]
        atom = pn.diags_dict[d][0]
        # split e.g. obtain 'O', 3 from 'O3'
        elem, spec = pn.utils.misc.parseAtom(atom)
        # instantiate the EmisGrid object for the given ion 
        EG = pn.EmisGrid(elem, spec)
        # select the axis in which to plot
        ax = axes.ravel()[i]
        # make the plot
        EG.plotContours(to_eval = to_eval, ax=ax)
   
    plt.close()

def test_line_labels():
    """
    This tool scans all the lines from pn.LINE_LABEL_LIST, extract the wavelength and compare with the 
    wavelength given by the corresponding atom for the corresponding transition. Is the difference is greater 
    than 1AA for optical or 0.1mu for IR, it prints the information
    """
    for atom_str in pn.LINE_LABEL_LIST:
        try:
            if atom_str[0] == '3':
                atom = None
            elif atom_str[-1] == 'r':
                atom = pn.RecAtom(atom=atom_str)
            else:
                atom = pn.Atom(atom=atom_str)
        except:
            atom = None
        if atom is not None:
            if atom.NLevels != 0:
                for line in pn.LINE_LABEL_LIST[atom_str]:
                    if line[-1] == 'm':
                        wavelength = float(line[:-1])*1e4
                    else:
                        wavelength = float(line[:-1])
                    i, j = atom.getTransition(wavelength)
                    wave = atom.wave_Ang[i-1, j-1]
                    if wave < 1e4:
                        dif = np.abs(wavelength-wave)
                    else:
                        dif = np.abs(wavelength-wave)/1e3
                    if dif > 1.0:
                        print(atom, line)
                        assert dif < 1.0

def test_atom_instanciation():
    """
    This tool try to instantiate all the atoms using all the atomic data.
    """
    atoms = pn.atomicData.getAllAtoms()
    atoms.remove('3He2')
    for atom in atoms:
        for atom_file in pn.atomicData.getAllAvailableFiles(atom, 'atom'):
            for coll_file in pn.atomicData.getAllAvailableFiles(atom, 'coll'):
                #print atom, atom_file, coll_file
                try:
                    pn.atomicData.setDataFile(atom_file)
                    pn.atomicData.setDataFile(coll_file)
                    atm = pn.Atom(atom=atom)
                    assert atm.atom == atom
                except:
                    assert None == atom
        for rec_file in pn.atomicData.getAllAvailableFiles(atom, 'rec'):
                try:
                    pn.atomicData.setDataFile(rec_file)
                    atm = pn.Atom(atom=atom)
                    assert atm.atom == atom
                except:
                    assert None == atom
            
def test_atom_props():
    
    O3 = pn.Atom("O", 3)
    
    assert O3.elem == "O"
    assert O3.spec == 3
    assert O3.atom == "O3"
    assert O3.getTransition(5007) == (4, 3)
    
def test_atom_emissivity_values(error=5e-2):
    
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_SSB14.dat')
    atom = pn.Atom("O", 3)

    result = atom.getEmissivity(20000, 1e2)
    emi = [9.55e-22, 4.00e-28, 7.77e-22, 1.42e-24, 3.98e-21, 1.19e-20, 9.49e-23, 
        2.78e-25, 4.01e-22]
    a = []
    for i in result:
        for j in i:
            if (j > 0) :
                a.append(j)

    for aux in range(len(a) - 1):
        result = (a[aux] / emi[aux])
        assert (rel_error(a[aux], emi[aux]) < error)
    
def test_atom_emissivity_shape():
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_AK99.dat')
    atom = pn.Atom("O", 3)
    result = atom.getEmissivity(20000, 1e2)
    assert (result.shape == (6, 6))
    result = atom.getEmissivity(20000, [1e2, 1e3])
    assert (result.shape == (6, 6, 2))
    result = atom.getEmissivity([20000, 10000], 1e2)
    assert (result.shape == (6, 6, 2))
    result = atom.getEmissivity([20000, 10000], [1e2, 1e3])        
    assert (result.shape == (6, 6, 2, 2))
    result = atom.getEmissivity([20000, 10000], [1e2, 1e3], product = True)        
    assert (result.shape == (6, 6, 2, 2))
    result = atom.getEmissivity([20000, 10000], [1e2, 1e3], product = False)        
    assert (result is None)
    
    result = atom.getEmissivity(20000, 1e2, wave = 5007)
    assert (result.shape == ())
    result = atom.getEmissivity(20000, [1e2, 1e3], wave = 5007)
    assert (result.shape == (2,))
    result = atom.getEmissivity([20000, 10000], 1e2, wave = 5007)
    assert (result.shape == (2,)) # the bug I'm looking for
    result = atom.getEmissivity([20000, 10000], [1e2, 1e3], wave = 5007)        
    assert (result.shape == (2, 2))
    result = atom.getEmissivity([20000, 10000], [1e2, 1e3], product = True, wave = 5007)        
    assert (result.shape == (2, 2))
    
def test_HbEmissivity(max_error=5e-4):
    H1 = pn.RecAtom('H', 1)
    tem = np.array([.5, 1, 1.5, 2., 3.]) * 1e4
    den = 1e2
    Hb_expected = np.array([2.2e-25, 1.235e-25, 8.6e-26, 6.58e-26, 4.440e-26])
    Hb_computed = H1.getEmissivity(tem, den, label='4_2')
    for i in range(len(tem)):
        assert(rel_error(H1.getEmissivity(tem[i], den, label='4_2'), Hb_expected[i]) < max_error)
        assert(rel_error((Hb_computed[i]), Hb_expected[i]) < max_error)
    
def test_ion_abundance(max_error=5e-3):
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_AK99.dat')
    atom = pn.Atom("O", 3)
    assert(rel_error(atom.getIonAbundance(10000, 10000, 100, 4, 3), 3.62e-3) < max_error)
    assert(rel_error(atom.getIonAbundance(10000, 20000, 100, 4, 3), 5.662e-4) < max_error)
    assert(rel_error(atom.getIonAbundance(10000, 10000, 100, wave=5007), 3.62e-3) < max_error)
    assert(rel_error(atom.getIonAbundance(10000, 20000, 100, wave=5007), 5.662e-4) < max_error)
    pn.atomicData.setDataFile('s_ii_atom_PKW09.dat')
    pn.atomicData.setDataFile('s_ii_coll_TZ10.dat')
    atom = pn.Atom("S", 2)
    assert(rel_error(atom.getIonAbundance(200, 20000, 100, 3, 1), 2.295e-6) < max_error)
    pn.atomicData.setDataFile('o_ii_atom_Z82-WFD96.dat')
    pn.atomicData.setDataFile('o_ii_coll_P06-T07.dat')
    atom = pn.Atom("O", 2)
    assert(rel_error(atom.getIonAbundance(200, 20000, 1000, 3, 1), 1.548e-5) < max_error)

def test_getA(max_error=5e-3):
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_AK99.dat')
    atom = pn.Atom("O", 3)
    assert(rel_error(atom.getA(1, 1) + 1, 1.0) < max_error)
    assert(rel_error(atom.getA(2, 1), 2.62e-05) < max_error)
    assert(rel_error(atom.getA(3, 1), 3.17e-11) < max_error)
    assert(rel_error(atom.getA(4, 1), 2.41e-06) < max_error)
    assert(rel_error(atom.getA(3, 2), 9.76e-05) < max_error)
    assert(rel_error(atom.getA(4, 2), 0.0068) < max_error)
    assert(rel_error(atom.getA(5, 2), 0.215) < max_error)

def test_getOmega(max_error=5e-3):
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_AK99.dat')
    atom = pn.Atom("O", 3)
    assert(rel_error(atom.getOmega(10000, 2, 1), 0.522) < max_error)
    assert(rel_error(atom.getOmega(10000, 3, 1), 0.2573) < max_error)
    assert(rel_error(atom.getOmega(10000, 4, 1), 0.2434) < max_error)
    assert(rel_error(atom.getOmega(10000, 5, 1), 0.0321) < max_error)
    assert(rel_error(atom.getOmega(10000, 3, 2), 1.232) < max_error)
    assert(rel_error(atom.getOmega(10000, 4, 2), 0.73) < max_error)
    assert(rel_error(atom.getOmega(10000, 5, 2), 0.0962) < max_error)
    assert(rel_error(atom.getOmega(10000, wave=5007), 1.217) < max_error)
    
def test_getTemDen(max_error=5e-3):
    pn.atomicData.setDataFile('o_iii_atom_SZ00-WFD96.dat')
    pn.atomicData.setDataFile('o_iii_coll_AK99.dat')
    atom = pn.Atom('O', 3)
    assert(rel_error(atom.getTemDen(100., den=10000., wave1=5007, wave2=4363), 11308.) < max_error)
    assert(rel_error(atom.getTemDen(1., tem=10000., wave1=88e4, wave2=52e4), 183) < max_error)
    pn.atomicData.setDataFile('o_ii_atom_Z82-WFD96.dat')
    pn.atomicData.setDataFile('o_ii_coll_P06-T07.dat')
    atom = pn.Atom('O', 2)
    assert(rel_error(atom.getTemDen(1.5, tem=10000., wave1=3726, wave2=3729), 1496) < max_error)

                   


import numpy as np
from scipy import interpolate
import os
import sys
import re
import pyneb as pn
if pn.config.Chianti_version_main == '8':
    from . import _chianti_tools_8 as _chianti_tools
elif pn.config.Chianti_version_main == '7':
    from . import _chianti_tools
elif pn.config.Chianti_version_main == '9':
    from . import _chianti_tools_9 as _chianti_tools
from . import _chianti_constants as const
from .physics import sym2name, vactoair
from .manage_atomic_data import getLevelsNIST, atom2chianti
from .misc import parseAtom

def Chianti_getA(ion_chianti, NLevels=None):
    """
    Return As from Chianti database.
    Usage:
        As_O3 = Chianti_getA('o_3')
    """
    wgfa = _chianti_tools.wgfaRead(ion_chianti)
    NLevels_ori = np.max(wgfa['lvl2'])
    As = np.zeros((NLevels_ori, NLevels_ori))
    for i, j, a in zip(wgfa['lvl2'], wgfa['lvl1'], wgfa['avalue']):
        As[i-1, j-1] = a
    if NLevels is not None:
        As = As[0:NLevels, 0:NLevels]
    return As

def Chianti_getE(ion_chianti, NLevels=None):
    """
    Return energies from the Chianti database
    Usage:
        E_O3 = Chianti_getE('o_3')
    """
    
    Es = _chianti_tools.elvlcRead(ion_chianti)
    if not 'j' in Es:
        return None
    if sys.version_info.major < 3:
        dtype = [('ecmth', '<f8'), ('ionS', 'S10'), ('term', 'S20'),  ('pretty', 'S30'), 
                 ('spd', 'S10'), ('ecm', '<f8'), ('j', '<f8'), ('l', '<i8'), ('erydth', '<f8'), 
                 ('conf', '<i8'), ('lvl', '<i8'), ('spin', '<i8'), ('eryd', '<f8'), ('mult', '<i8')]
    else:
        dtype = [('ecmth', '<f8'), ('ionS', 'U10'), ('term', 'U20'),  ('pretty', 'U30'), 
                 ('spd', 'U10'), ('ecm', '<f8'), ('j', '<f8'), ('l', '<i8'), ('erydth', '<f8'), 
                 ('conf', '<i8'), ('lvl', '<i8'), ('spin', '<i8'), ('eryd', '<f8'), ('mult', '<i8')]        
    arr = np.recarray((len(Es['j']),), dtype=dtype)
    for k in Es.keys():
        if k not in ('status', 'ref', 'label', 'filename'):
            arr[k] = Es[k]
    if NLevels is not None:
        arr = arr[0:NLevels]
    return arr

def get_levs_order(atom, NLevels=None):
    """
    Returns a dictionary giving the correspondence of indices between NIST and Chianti energy level tables.
    
    """
    E_chianti = Chianti_getE(atom2chianti(atom))
    if E_chianti is None:
        return None
    E_NIST = getLevelsNIST(atom)
    if E_NIST is None:
        return None
    # Here follow a list of filter to try to find coincidence between the configuration used in Chanti and in NIST
    remove_stars = lambda str: re.sub('\*','', str)
    remove_par = lambda str: re.sub(r'\([^)]*\)','', str)
    change_2pto1p = lambda str: re.sub('\.\.', '.', str)
    change_2pto0p = lambda str: re.sub('\.\.', '', str)
    remove_point_before_par = lambda str: re.sub('\.\(', '(', str)
    remove_before_firstp = lambda str: re.sub('^[^.]*.', '', str)
    change_ptpsp = lambda str: re.sub('\.', ' ', str)
    # Define the Dictionary
    Chianti2NIST = {}
    if NLevels is not None:
        this_NLevels = min((NLevels, len(E_chianti), len(E_NIST)))
    else:
        this_NLevels = min((len(E_chianti), len(E_NIST)))
    for i_N, E in enumerate(E_NIST):
        
        if i_N >= len(E_chianti):
            break
        if i_N > this_NLevels:
            break
        
        term = remove_stars(E['term'])

        c1 = remove_stars(E['conf'])
        
        c12 = remove_par(c1)
        c123 = change_2pto1p(c12)
        c124 = change_2pto0p(c12)
        
        c15 = remove_point_before_par(c1) 
        
        c16 = remove_before_firstp(c1)
        c126 = remove_before_firstp(c12)
        c1236 = remove_before_firstp(c123)
        c1246 = remove_before_firstp(c124)
        c156 = remove_before_firstp(c15)
        
        c17 = change_ptpsp(c1)
        c127 = change_ptpsp(c12)
        c1237 = change_ptpsp(c123)
        c1247 = change_ptpsp(c124)
        c157 = change_ptpsp(c15)
        c167 = change_ptpsp(c16)
        c1267 = change_ptpsp(c126)
        c12367 = change_ptpsp(c1236)
        c12467 = change_ptpsp(c1246)
        c1567 = change_ptpsp(c156)
       
        confs = (c1, c12, c123, c124, c15, c16, c126, c1236, c1246, c156,
                 c17, c127, c1237, c1247, c157, c167, c1267, c12367, c12467, c1567)
        for conf in confs:   
            pretty = '{0} {1}{2}'.format(conf, term, E['J'])
            i_Ch = np.where(E_chianti['pretty'] == pretty)[0]
            if len(i_Ch) == 1:
                if i_Ch[0] <= this_NLevels and i_N <= this_NLevels:
                    Chianti2NIST[i_N] = i_Ch[0]
    if len(Chianti2NIST) == 0:
        Chianti2NIST = None
        
        ### Il faut que ni i_N ni i_Ch depasse NLvels. Et il faut que NLevels soit diminue dansla cas contraire.
    return Chianti2NIST
        
def Chianti_getOmega(ion_chianti, tem, lev1=None, lev2=None, Splups=None, NLevels=None):
    """
    Return the values of Upsilon at a given temperature (may be a table)
    Usage:
        Chianti_getOmega('o_3', 10000)
        Chianti_getOmega('o_3', np.linspace(5000, 20000, 20))
    lower and upper levels can be given as lev1 and lev2. If not, full table is return
    Splups is the result of _chianti_tools.splupsRead(ion_chianti), can be passed to avoid reloading it
    NLevelsMax reduce the number of levels considered.
    """
    pn.log_.debug('NLevels = {}'.format(NLevels), calling='Chianti_getOmega')
    temp=np.asarray(tem)
    if Splups is None:
        if pn.config.Chianti_version_main == '9':
            Splups = _chianti_tools.scupsRead(ion_chianti)
        elif pn.config.Chianti_version_main == '8':
            Splups = _chianti_tools.scupsRead(ion_chianti)
        elif pn.config.Chianti_version_main == '7':
            Splups = _chianti_tools.splupsRead(ion_chianti)
    if NLevels is None:
        nsplups = len(Splups["lvl1"])
        nlevels = np.max(Splups["lvl2"])
    else:
        nsplups = np.min((len(Splups["lvl1"]), NLevels))
        nlevels = np.min((np.max(Splups["lvl2"]), NLevels))
    ntemp=temp.size
    if (lev1 is None) and (lev2 is None):
        Omega = np.zeros((nlevels, nlevels, ntemp))
    else:
        Omega = np.zeros(ntemp)
    lvl1 = np.array(Splups["lvl1"])
    lvl2 = np.array(Splups["lvl2"])
    if lev1 is None:
        range1 = np.arange(nlevels)
    else:
        range1 = np.array([lev1-1])
    if lev2 is None:
        range2 = np.arange(nlevels)
    else:
        range2 = np.array([lev2-1])
    for l1 in range1:
        for l2 in range2:
            isplups = np.where(((lvl1 == (l1+1)) & (lvl2 == (l2+1))))
            if type(isplups) == type(()):
                isplups = isplups[0]                
            if len(isplups) == 0:
                ups = np.zeros(ntemp)
            else:
                if len(isplups) > 1 or np.ndim(isplups)==1:
                    isplups = isplups[0]
                ttype=Splups["ttype"][isplups]
                cups=Splups["cups"][isplups]
                if pn.config.Chianti_version_main == '9':
                    nspl=Splups["ntemp"][isplups]
                    dx=1./(float(nspl)-1.)
                    xs = Splups['btemp'][isplups]
                    splups=Splups["bscups"][isplups]
                elif pn.config.Chianti_version_main == '8':
                    nspl=Splups["ntemp"][isplups]
                    dx=1./(float(nspl)-1.)
                    xs = Splups['btemp'][isplups]
                    splups=Splups["bscups"][isplups]
                elif pn.config.Chianti_version_main == '7':
                    nspl=Splups["nspl"][isplups]
                    dx=1./(float(nspl)-1.)
                    xs=dx*np.arange(nspl)
                    splups=Splups["splups"][isplups]
                de=Splups['de'][isplups]
                kte = const.boltzmann*temp/(de*const.ryd2erg)
                der=0
                if ttype == 1:
                    st=1.-np.log(cups)/np.log(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=sups*np.log(kte+np.exp(1.))
                elif ttype == 2:
                    st=kte/(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=sups
                elif ttype == 3:
                    st=kte/(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=sups/(kte+1.)
                elif ttype == 4:
                    st=1.-np.log(cups)/np.log(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=sups*np.log(kte+cups)
                elif ttype == 5:
                    # dielectronic rates
                    st=kte/(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=sups/(kte+0.)
                elif ttype == 6:
                    st=kte/(kte+cups)
                    y2=interpolate.splrep(xs,splups,s=0)
                    sups=interpolate.splev(st,y2,der=der)
                    ups=10.**sups
                elif ttype > 6:
                    pn.log_.warning(' t_type ne 1,2,3,4,5={} {} {}'.format(ttype,l1,l2), calling='pn.chianti.Chianti_getOmega')
            if (lev1 is None) and (lev2 is None):
                Omega[l2, l1, :] = ups
            else:
                Omega = ups

    Omega = np.where(Omega > 0.,Omega,0.)
    return Omega

class _AtomChianti(object):
    
    def __init__(self, elem=None, spec=None, atom=None, NLevels=None):
        """
        Object dealing with As values from the Chianti database.
        The directory where to find the data must be given through the environment variable XUVTOP.
        This object is not aimed to be used by itself, it is called by Atom.
        """
        self.log_ = pn.log_
        if atom is not None:
            self.atom = atom
            self.elem = parseAtom(atom)[0]
            self.spec = int(parseAtom(atom)[1])
        else:
            self.elem = elem
            self.spec = int(spec)
            self.atom = elem + str(self.spec)
        self.name = sym2name[self.elem]
        self.ion_chianti =  atom2chianti(self.atom)
        self.calling = 'Atom {}'.format(self.atom)
        self.Chianti2NIST = None
        self.NLevels = NLevels
        self._loadChianti()
        self.initWaves()
        
    def _loadChianti(self):
        self.fullFileName = _chianti_tools.ion2filename(self.ion_chianti) + '.wgfa'
        
        self.atomFile = self.fullFileName.split('/')[-1]
        self.atomPath = self.fullFileName[:-len(self.atomFile)]
        
        if not os.path.exists(self.fullFileName):
            self.log_.error('File {0} not found'.format(self.fullFileName), calling=self.calling)
        
        self.Chianti_version = self.atomPath.split('/')[-4]
        self.comments = {}
        
        if self.Chianti2NIST is None:
            self.Chianti2NIST = get_levs_order(self.atom, NLevels=self.NLevels)

        if self.NLevels is None:
            this_NLevels = None
        else:
            if self.Chianti2NIST is not None:
                this_NLevels = np.max((self.NLevels, np.max(list(self.Chianti2NIST.values())[0:self.NLevels])))
            else:
                this_NLevels = self.NLevels
                
        Chianti_A = Chianti_getA(self.ion_chianti, NLevels=this_NLevels)
        if self.Chianti2NIST is not None:
            Chianti_A_tmp = Chianti_A.copy()
            for i_chianti in self.Chianti2NIST:
                if self.Chianti2NIST[i_chianti] != i_chianti:
                    Chianti_A[i_chianti,:] = Chianti_A_tmp[self.Chianti2NIST[i_chianti],:]
                    Chianti_A[:,i_chianti] = Chianti_A_tmp[:,self.Chianti2NIST[i_chianti]]
        if self.NLevels is not None:
            Chianti_A = Chianti_A[0:self.NLevels, 0:self.NLevels]
        self.log_.message('Reading atom data from Chianti {}'.format(self.atomFile), calling = self.calling)
        self.Chianti_E = Chianti_getE(self.ion_chianti, NLevels=self.NLevels)
        if self.NLevels is None:
            self.NLevels = len(self.Chianti_E)
        self.NIST = getLevelsNIST(self.atom, self.NLevels)
        need_NIST = True
        
        if self.NIST is not None:
            self.NLevels = np.min((len(self.NIST), self.NLevels))    
            energy = self.NIST['energy'] / 1e8
            stat_weight = 1 + 2 * self.NIST['J']
            self.comments['VACUUM'] = '1'
            self.comments['NOTE'] = 'Energy levels'
            self.comments['SOURCE'] = 'NIST 2014'
        elif need_NIST:
            pn.log_.error('NIST data are needed for this format of atomic data', calling=self.calling) 

        self.E_in_vacuum = True
        self._Energy = energy
        self._StatWeight = stat_weight
        self._A = Chianti_A
        self.atomNLevels = self.NLevels
            
    def initWaves(self):
        """
        Initialization of wave_Ang
        
        """
        self.wave_Ang = np.zeros((self.NLevels, self.NLevels))
        
        for i in range(1, self.NLevels):
            for j in range(i):
                wave = 1. / abs(self._Energy[i] - self._Energy[j])
                if self.E_in_vacuum:
                    wave = vactoair(wave)
                self.wave_Ang[i, j] = self.wave_Ang[j, i] = wave
 
    def _test_lev(self, level):
        """
        Test whether selected level is legal

        Parameters:
            - level        selected atom level

        """       
        if level < -1 or level == 0 or level > self.NLevels:
            self.log_.error('Wrong value for level: {0}, maximum = {1}'.format(level, self.NLevels),
                            calling=self.calling)
    
    def getA(self, lev_i= -1, lev_j= -1):
        """
        Return the transition probability data. 
        If no arguments are given, the whole array of A is returned.
        A specific A value can be obtained by giving either the upper and lower levels.
            
        Usage:
            A_O3 = O3.getA()          # The whole A array is stored in A_O3
            O3.getA(4, 2)      # A(4, 2) of the O3 atom is printed
            O3.getA(2, 4)      # Returns 0

        Parameters:
            - lev_i  upper level of transition (default= -1, returns complete array)
            - lev_j  lower level of transition (default= -1, returns complete array)
            
        """
        
        self._test_lev(lev_i)
        self._test_lev(lev_j)
        if (lev_i == -1):
            if (lev_j == -1):
                return self._A
            else:
                return (self._A[lev_i - 1])
        else:
            return (self._A[lev_i - 1, lev_j - 1])

    def getStatWeight(self, level= -1):
        """
        Returns the array of statistical weights of the ion (if no arguments are given) 
            or the statistical weight of level i (if i is given as an argument, 
            with the optional keyword level).
            
        Usage:
            O3.getStatWeight()
            O3.getStatWeight(level=4)
            O3.getStatWeight(4)
                      
        Parameters:
            - level  selected atomic level (default= -1, returns complete array)
            
        """
        self._test_lev(level)
        
        if level == -1:
            return self._StatWeight
        else:
            return self._StatWeight[level-1]
        
    
    def getEnergy(self, level= -1, unit='1/Ang'):
        """
        Return energy level of selected level (or array of energy levels, if level not given) 
            in Angstrom^-1 (default) or another unit
        
        Usage:
            O3.getEnergy(4, unit='eV')
        Parameters:
            - level  selected atomic level (default= -1, returns complete array)
            - unit   [str] one of '1/Ang' (default), 'eV', or 'Ryd'    
            
        """
        self._test_lev(level)

        unit_dict = {'1/Ang': 1.,
                     'Ryd': pn.CST.RYD_ANG,
                     'eV': pn.CST.RYD_ANG * pn.CST.RYD_EV,
                     'cm-1': 1e8}
        if unit not in unit_dict:
            self.log_.warn('Unit {0} unknown, using 1/Ang'.format(unit), calling=self.calling + '.getEnergy')
            unit = '1/Ang'        
        
        if level == -1:
            return self._Energy * unit_dict[unit]
        else:
            return self._Energy[level-1] * unit_dict[unit]
        
    def getSources(self):
        """
        Return bibliographic sources for atomic data, as listed in the headers of the fits files
        
        """
        sources = []
        sources.append('A-values from {0}'.format(self.Chianti_version))
        for ref in  _chianti_tools.elvlcRead(self.ion_chianti)['ref']:
            sources.append(ref)
        return sources

    def printSources(self):
        """
        Print bibliographic sources for atomic data, as listed in the headers of the fits files
        
        """
        print('A-values from {0}'.format(self.Chianti_version))
        for ref in  _chianti_tools.elvlcRead(self.ion_chianti)['ref']:
            print(ref)
               
class _CollChianti(object):
    
    def __init__(self, elem=None, spec=None, atom=None, NLevels=None, TemArray=np.logspace(2, 5, 20)):
        """
        Object dealing with Upsilon values from the Chianti database.
        The directory where to find the data must be given through the environment variable XUVTOP.
        This object is not aimed to be used by itself, it is called by Atom.
        """

        self.log_ = pn.log_
        
        if atom is not None:
            self.atom = atom
            self.elem = parseAtom(atom)[0]
            self.spec = int(parseAtom(atom)[1])
        else:
            self.elem = elem
            self.spec = int(spec)
            self.atom = elem + str(self.spec)
        self.name = sym2name[self.elem]
        self.ion_chianti =  atom2chianti(self.atom)
        self.calling = 'Atom ' + self.atom
        self.NLevels = NLevels
        self.tem_units = 'K'
        self._TemArray = TemArray
        self._loadChianti()
        
    def _loadChianti(self):
        if pn.config.Chianti_version_main == '9':
            self.fullFileName = _chianti_tools.ion2filename(self.ion_chianti) + '.scups'
        elif pn.config.Chianti_version_main == '8':
            self.fullFileName = _chianti_tools.ion2filename(self.ion_chianti) + '.scups'
        elif pn.config.Chianti_version_main == '7':
            self.fullFileName = _chianti_tools.ion2filename(self.ion_chianti) + '.splups'
        self.collFile = self.fullFileName.split('/')[-1]
        self.collPath = self.fullFileName[:-len(self.collFile)]
        
        if not os.path.exists(self.fullFileName):
            self.log_.error('File {0} not found'.format(self.fullFileName), calling=self.calling)
        
        self.Chianti_version = self.collPath.split('/')[-4]
        self.comments = {}
        if pn.config.Chianti_version_main == '9':
            self.Splups = _chianti_tools.scupsRead(self.ion_chianti)
        elif pn.config.Chianti_version_main == '8':
            self.Splups = _chianti_tools.scupsRead(self.ion_chianti)
        elif pn.config.Chianti_version_main == '7':
            self.Splups = _chianti_tools.splupsRead(self.ion_chianti)
        self._CollArray = Chianti_getOmega(self.ion_chianti, tem=self._TemArray, 
                                   Splups=self.Splups, NLevels=self.NLevels)
        to_minimize = (self._CollArray.shape[0]-1, self._CollArray.shape[1]-1, self.NLevels)
        if None not in to_minimize:    
                self.NLevels = np.min(to_minimize)
        self.Chianti2NIST = get_levs_order(self.atom, NLevels=self.NLevels)
        
        if self.Chianti2NIST is not None:
            _CollArray_tmp = self._CollArray.copy()
            for i_chianti in self.Chianti2NIST:
                if self.Chianti2NIST[i_chianti] != i_chianti:
                    self._CollArray[i_chianti,:,:] = _CollArray_tmp[self.Chianti2NIST[i_chianti],:,:]
                    self._CollArray[:,i_chianti,:] = _CollArray_tmp[:,self.Chianti2NIST[i_chianti],:]
        
        self.log_.message('Reading coll data from Chianti {}'.format(self.collFile), calling = self.calling)
        self.NLevels = self._CollArray.shape[0]


    def _test_lev(self, level):
        """
        Test whether selected level is legal

        Parameters:
            - level        selected atom level

        """       
        if level < -1 or level == 0 or level > self.NLevels:
            self.log_.error('Wrong value for level: {0} (maximum: {1})'.format(level, self.NLevels),
                            calling=self.calling)

    def getOmegaArray(self, lev_i= -1, lev_j= -1):
        """
        Return array of original tabulated collision strengths for a given transition, 
            as a function of temperature.
        
        Usage:
            O3.getOmegaArray()
        
        Parameters:
            - lev_j  lower level (default= -1, returns complete array)
            - lev_i  upper level (default= -1, returns complete array)

        """
        self._test_lev(lev_i)
        self._test_lev(lev_j)
        if (lev_i <= lev_j) and (lev_i != -1):
            self.log_.warn("wrong levels given {0} <= {1}".format(lev_i, lev_j), calling=self.calling)
            return None
        elif lev_i == -1 or lev_j == -1:
            return self._CollArray
        else:
            return self._CollArray[lev_i-1, lev_j-1,:]

    def getOmega(self, tem, lev_i= -1, lev_j= -1):
        """
        Return interpolated value of the collision strength value at the given temperature 
            for the complete array or a specified transition.

        Usage:
            O3.getOmega(15000.)
            O3.getOmega([8e3, 1e4, 1.2e4])
            O3.getOmega([8e3, 1e4, 1.2e4], 5, 4)
        
        Parameters:
            - tem    electronic temperature in K. May be an array.
            - lev_i  upper level
            - lev_j  lower level

        """
        self._test_lev(lev_i)
        self._test_lev(lev_j)
        if (lev_i <= lev_j) and (lev_i != -1):
            self.log_.warn("wrong levels given {0} <= {1}".format(lev_i, lev_j), calling=self.calling)
            return None            
        if self.Chianti2NIST is None:
            self.Chianti2NIST = get_levs_order(self.atom)
        if (lev_i == -1) and (lev_j == -1):
            Omega = Chianti_getOmega(self.ion_chianti, tem, Splups=self.Splups, NLevels=self.NLevels)
            if self.Chianti2NIST is not None:
                Omega_tmp = Omega.copy()
                for i_chianti in self.Chianti2NIST:
                    if self.Chianti2NIST[i_chianti] != i_chianti:
                        Omega[i_chianti,:,:] = Omega_tmp[self.Chianti2NIST[i_chianti],:,:]
                        Omega[:,i_chianti,:] = Omega_tmp[:,self.Chianti2NIST[i_chianti],:]
        else:
            if self.Chianti2NIST is not None:
                try:
                    Omega = Chianti_getOmega(self.ion_chianti, tem, lev1=self.Chianti2NIST[lev_j-1]+1, lev2=self.Chianti2NIST[lev_i-1]+1, 
                                             Splups=self.Splups, NLevels=self.NLevels)
                except:
                    Omega = 0.
            else:
                Omega = Chianti_getOmega(self.ion_chianti, tem, lev1=lev_j, lev2=lev_i, 
                             Splups=self.Splups, NLevels=self.NLevels)
                
        return np.squeeze(Omega)

    def getTemArray(self, keep_unit=True):
        """
        Return array of tabulated original temperature points (as in fits file) 
            of collision strengths.
        
        Parameters:
            - keep_unit   return temperature in file units (default) or change it to Kelvin (False)
        """
        if keep_unit:
            return self._TemArray
        else:            
            if (self.tem_units == "log(K)"):
                return pow(10., self._TemArray)
            else:
                if (self.tem_units == "K/10000"):
                    return self._TemArray * 1.e4
                else: 
                    return self._TemArray
    
    def getSources(self):
        """
        Return bibliographic sources for atomic data, as listed in the headers of the fits files
        
        """
        sources = []
        sources.append('Omega-values from {0}'.format(self.Chianti_version))
        
        if pn.config.Chianti_version_main == '9':
            refs = _chianti_tools.scupsRead(self.ion_chianti)
        elif pn.config.Chianti_version_main == '8':
            refs = _chianti_tools.scupsRead(self.ion_chianti)
        elif pn.config.Chianti_version_main == '7':
            refs = _chianti_tools.splupsRead(self.ion_chianti)
        for ref in  refs['ref']:
            sources.append(ref)
        return sources

    def printSources(self):
        
        for source in self.getSources():
            print(source)    
        
